# Paravirtualized guest support
#
# Initially derived from virt-inst, Copyright 2005-2006 Red Hat, Inc.
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import os
import re

from VMDefaults import VMDefaults
from xen_guest import *
from exceptions import *
from globals import *
from log import log
import mount
import paths
import util
import vmdisks
import time

def loopsetup(vdev):
    """Sets up the loop mapping for a disk image.
    """
    i = 0
    while True:
        ldev = '/dev/loop%i' % (i)
        if not os.path.exists(ldev):
            break
        i += 1
        fd = os.popen("losetup %s '%s' 2> /dev/null" % (ldev, vdev))
        if not fd.close():
            log.debug("losetup %s '%s'" % (ldev, vdev))
            return ldev
    return None

def loopclean(ldev):
    """Delete the loop mapping.
    """
    if ldev:
        log.debug("losetup -d %s" % ldev)
        retries = 10
        while retries:
            fd = os.popen("losetup -d %s" % ldev)
            if not fd.close():
                break
            else:
                time.sleep(0.1)
                retries -= 1

def _get_partitions(dev):
    """Returns a list of dictionaries that describe the (relevant)
    partitions on the device.
    """
    found = False
    partitions = []
    pre = re.compile('([^ ]*)[ \\t]*:[ \\t]*start=[ \\t]*([0-9]*)[ \\t]*, size=[ \\t]*([0-9]*)[ \\t]*, Id=[ \\t]*([0-9a-fA-F]*)')
    log.debug("Examining '%s' for partitions..." % dev)
    lines = util.run(['/sbin/sfdisk', '--force', '-d', '-uS', '--', dev])
    for line in lines.splitlines():
        m = pre.match(line)
        if m:
            partition = {}
            pname = m.group(1)
            pstart = m.group(2)
            pid = m.group(4)
            # For now, only investigate 83 (Linux) and 6 (FAT16)
            if pid not in ('6', '83'):
                continue
            log.debug("Found partition '%s'." % (pname,))
            partition['name'] = pname
            partition['id'] = pid
            partition['offset'] = int(pstart) * 512
            partitions.append(partition)
            found = True

    # Try '8e' (Linux LVM) scan
    if not found:
        log.debug("Try LVM scan: vdev %s" % (dev,))
        vgname = []
        fd = os.popen("pvscan | awk '$2 == \"%s\" {print $4}' 2>/dev/null" % (dev))
        for line in fd.readlines():
            line = line.strip()
            vgname = line[:]
        fd.close()
        if not vgname:
            return partitions

        log.debug("LVM Found vg: vdev %s" % (vgname))
        ret = os.system("vgchange -ay '%s' > /dev/null 2>&1" % (vgname)) >> 8
        if not ret:
            log.debug("lvscan | grep '/dev/%s'" % (vgname))
            fd = os.popen("lvscan | grep '/dev/%s'" % (vgname))
            for line in fd.readlines():
                line = line.strip()
                (t1, lvname, t2) = line.split('\'')
                pname = lvname[lvname.rfind('/')+1:]
                pname = pname.strip()
                pname = "/dev/mapper/" + vgname + "-" + pname
                partition = {}
                partition['name'] = pname
                partition['id'] = '8e'
                partition['offset'] = 0
                partition['vdev'] = vgname
                partitions.append(partition)
                log.debug("Found partition: vdev %s, pdev %s" % (vgname, pname))
            fd.close()
            log.debug("vgchange -an '%s'" % (vgname))
            os.system("vgchange -an '%s' > /dev/null 2>&1" % (vgname))
        
    return partitions

class ParaVirtGuest(XenGuest):
    def __init__(self, options):
        # Set these early, in case the XenGuest.__init__ raises.
        self.kernel = ''
        self.initrd = ''
        XenGuest.__init__(self, options)
        self.runtime = {}

    def _get_paravirt_images(self, dev, partitions, copy, progress=lambda f:None):
        def traildigits(strg):
            """Return the trailing digits, used to split the partition number off"""
            idx = len(strg)-1
            while strg[idx].isdigit():
                if len == 0:
                    return strg
                idx -= 1
            return strg[idx+1:]
        actions = VMDefaults[self.options.os_type].actions()
        for partition in partitions:
            # For now, only investigate 6 (FAT16) and 83 (Linux), 8e (Linux LVM)
            if partition['id'] not in ('6', '83', '8e'):
                continue
            if partition['id'] == '8e':
                os.system("vgchange -ay '%s' > /dev/null 2>&1" % (partition['vdev']))
                dev = partition['name']
            log.debug("Loop-mounting '%s', offset %d..." % (dev, partition['offset']))
            vdisk = mount.TemporaryLoopMount(dev)
            try:
                mountpoint = vdisk.mount(options='loop,ro,offset=%d' % partition['offset'])
            except Exception, e:
                log.debug("Mount failed (%s); skipping." % str(e))
                if partition['id'] == '8e':
                    os.system("vgchange -an '%s' > /dev/null 2>&1" % (partition['vdev']))
                continue
            try:
                (_kernel, _initrd) = actions.getRuntimeImagesFromPath(mountpoint)
                if _kernel:
                    if partition['id'] == '8e':
                        vpart = 'LVM-' + partition['name'][partition['name'].rfind('/')+1:]
                    else:
                        vpart = traildigits(partition['name'])
                    log.info("Found kernel on partition %s." % vpart)
                    if copy:
                        # Be careful to explicitly close on failure, and close
                        # immediately after use, otherwise no guarantee these
                        # will be closed by GC before attempting unmount.
                        kfd = open(mountpoint + _kernel, 'r')
                        try:
                            ifd = None
                            if _initrd:
                                ifd = open(mountpoint + _initrd, 'r')
                        except:
                            kfd.close()
                            raise
                        try:
                            kernel = util.copy_to_temp(kfd, prefix='kernel.')
                            if not ifd:
                                initrd = None
                            else:
                                try:
                                    initrd = util.copy_to_temp(ifd, prefix='initrd.')
                                except:
                                    try: os.unlink(kernel)
                                    except: pass
                                    raise
                        finally:
                            kfd.close()
                            if ifd:
                                ifd.close()
                    else:
                        log.info("Not copying; will use kernel/initrd in-place")
                        kernel = _kernel
                        initrd = _initrd
                    log.info("kernel %s; initrd %s" % (kernel, initrd))
                    return (vpart, kernel, initrd)
            finally:
                vdisk.unmount()
                if partition['id'] == '8e':
                    log.debug("vgchange -an '%s'" % (partition['vdev']))
                    os.system("vgchange -an '%s' > /dev/null 2>&1" % (partition['vdev']))
                progress(1.0)
        progress(1.0)
        return (None, None, None)

    def _get_paravirt_install_images(self, progress=lambda f:None):
        """Retrieves the kernel and initrd from the installation source.
        Potentially a heavy-weight operation."""
        actions = VMDefaults[self.options.os_type].actions()
        source = self.options.source
        nonpae = VMDefaults[self.options.os_type].supports_32nonpae()
        if type(source) is vmdisks.Disk:
            log.debug("Installation source: %s" % source)
            if not source.exists():
                raise InstSrcError(InstSrcError.E_NOT_FOUND, details=str(source))
            vdev = source.block_attach()
            try:
                partitions = _get_partitions(vdev)
                if len(partitions) == 0:
                    mounted_disk = mount.TemporaryLoopMount(vdev)
                    try:
                        mountpoint = mounted_disk.mount()
                    except Exception, e:
                        raise InstSrcError(InstSrcError.E_BAD, details=e.args[0])
                    try:
                        (kfn, ifn) = actions.getInstallImagesFromPath(mountpoint, progress, nonpae)
                    finally:
                        mounted_disk.unmount()
                else:
                    (_, kfn, ifn) = self._get_paravirt_images(vdev, partitions, copy=True, progress=progress)
            finally:
                source.block_detach()
        else:
            (kfn, ifn) = actions.getInstallImagesFromUrl(self.options, source, progress, nonpae)
        if kfn is None:
            raise InstSrcError(InstSrcError.E_BAD, details=str(source))
        return (kfn, ifn)

    def _get_install_xml(self):
        os = "<os type='%s'" % self.options.os_type
        if self.options.install == OS_INSTALL_YES:
            if type(self.options.source) is vmdisks.Disk:
                src = 'dev:/' + self.options.source.get_vdev_str()
            else:
                src = self.options.source
            os += " source='%s'" % src
        os += '>'
        return [ os,
                 [ '<type>linux</type>',
                   '<kernel>%s</kernel>' % self.kernel,
                   '<initrd>%s</initrd>' % self.initrd,
                   '<cmdline>%s %s</cmdline>' % (self.options._install_args, self.options.extra_args) ],
                 '</os>' ]

    def _get_runtime_xml(self):
        ramdisk = ''
        if self.runtime['initrd']:
            ramdisk = ',%s' % self.runtime['initrd']
        return [ "<os type='%s'>" % self.options.os_type,
                 [ '<type>linux</type>',
                   '<bootloader>%s</bootloader>' % self.options.loader,
                   '<bootargs>--entry=%s:%s%s</bootargs>' % (self.runtime['vdev'], self.runtime['kernel'], ramdisk),
                   '<cmdline>%s %s</cmdline>' % (self.options._runtime_args, self.options.extra_args) ],
                 '</os>' ]

    def GetConfigXml(self):
        if self.isInstall:
            os_xml = self._get_install_xml()
        else:
            os_xml = self._get_runtime_xml()
            
        return [ "<domain type='xen'>",
                 self.getCommonXml(),
                 os_xml,
                 [ '<devices>',
                   [ self._get_disk_xml(),
                     self._get_network_xml(),
                     self._get_graphics_xml() ],
                   '</devices>' ],
                 '</domain>' ]
                 
    def _get_install_xen(self):
        config = 'builder="linux"\nkernel="%s"\n' % self.kernel
        if self.initrd:
            config += 'ramdisk="%s"\n' % self.initrd
        config += 'extra="%s %s"\n' % (self.options._install_args, self.options.extra_args)
        return config

    def _get_runtime_xen(self):
        ramdisk = ''
        if self.runtime['initrd']:
            ramdisk = ',%s' % self.runtime['initrd']
        if 'pygrub' in self.options.loader:
            return """builder="linux"
bootloader="%(loader)s"
bootargs=""
extra="%(runtime_args)s %(extra_args)s"
""" % { 'loader': self.options.loader, 'runtime_args': self.options._runtime_args, 'extra_args': self.options.extra_args }
        else:
            return """builder="linux"
bootloader="%(loader)s"
bootargs="--entry=%(vdev)s:%(kernel)s%(ramdisk)s"
extra="%(runtime_args)s %(extra_args)s"
""" % { 'loader': self.options.loader, 'vdev': self.runtime['vdev'], 'kernel': self.runtime['kernel'],
        'ramdisk': ramdisk, 'runtime_args': self.options._runtime_args, 'extra_args': self.options.extra_args }

    def GetConfigXen(self):
        if self.isInstall:
            osblob = self._get_install_xen()
        else:
            osblob = self._get_runtime_xen()

        return """%(common)s%(osblob)s%(disks)s
%(networks)s
%(graphics)s
""" % { 'common': self.getCommonXen(), 'osblob': osblob, 'disks': self._get_disk_xen(), 'networks': self._get_network_xen(), 'graphics': self._get_graphics_xen() }

    def PreInstall(self, progress=lambda f:None):
        if self.options.install != OS_INSTALL_UPGRADE:
            XenGuest.PreInstall(self, lambda f:progress(f*0.5))
        self._cleanup_kernel()  # In case the install was retried
        (self.kernel, self.initrd) = self._get_paravirt_install_images(lambda f:progress(0.5+f*0.5))

    def Start(self):
        try:
            XenGuest.Start(self)
        finally:
            self._cleanup_kernel()
        
    def PostInstallCheck(self):
        """Probes the installed disk(s) for kernel/initrd to boot from."""
        def _get_paravirt_runtime_images():
            r = (None, None, None)
            for disk in self.options.disks:
                # FIXME: skip installation source?
                try:
                    vdev = disk.block_attach()
                except:
                    continue
                try:
                    try:
                        log.debug("Post install check, loopsetup...")
                        ldev = loopsetup(vdev)
                        if ldev is not None:
                            partitions = _get_partitions(ldev)
                            r = self._get_paravirt_images(ldev, partitions, copy=False)
                            loopclean(ldev)
                        else:
                            partitions = _get_partitions(vdev)
                            r = self._get_paravirt_images(vdev, partitions, copy=False)
                    except OSError, ex:
                        log.debug("Probe kernel error:%d, %s" % (ex.errno, ex.strerror))
                        loopclean(ldev)
                        continue
                finally:
                    disk.block_detach()
                    if r[0] or r[1]:
                        break
            if r[0] is None or r[1] is None:
                log.debug("Raise Error No Kernel r[0]:%s, r[1]:%s" % (r[0], r[1]))
                raise VmError(VmError.E_NO_KERNEL)
            if r[0][:r[0].find('-')] == 'LVM':
                return (r[0], r[1], r[2])
            else:
                return (disk.get_vdev_str() + r[0], r[1], r[2])
        (vdev, kernel, initrd) = _get_paravirt_runtime_images()
        # Success:  No turning back to retry installation now!
        util.addInstallSourceRepo(self.options.source)
        self.runtime['vdev'] = vdev
        self.runtime['kernel'] = kernel
        self.runtime['initrd'] = initrd

    def _cleanup_kernel(self):
        if self.options.preserve_on_error:
            log.info("Installation failed, not cleaning up disks because preserve_on_error is set")
            self.kernel = ''
            self.initrd = ''
            return
        if self.kernel:
            try:
                os.unlink(self.kernel)
            finally:
                self.kernel = ''
        if self.initrd:
            try:
                os.unlink(self.initrd)
            finally:
                self.initrd = ''

    def __del__(self):
        log.debug("ParaVirtGuest.__del__")
        self._cleanup_kernel()
