# Functions used to manipulate advanced HVM guest options
#
# Copyright 2010 Novell, Inc.
# Author:
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import caps
from VMDefaults import VMDefaults
from exceptions import *

# The name and default value should match what is found in xen/xm/create.py
# List must be alphebetical
# Syntax: 'config option string': ['default value', "Description"]
advanced_xen_args = {
    'acpi': ['1', "Enable Advanced Configuration and Power Interface."],
    'apic': ['1', "Enable Advanced Programmable Interrupt Controller."],
    'blkif': ['0', "Make the domain a block device backend."],
    'hap': ['1', "Enable Hardware Assisted Paging"],
    'hpet': ['0', "Enable virtual High-Precision Event Timer."],
    'isa': ['0', "Simulate an Industry Standard Architecture only system"],
    'localtime': ['0', "Set Real-Time Clock (RTC) to localtime"],
    'memory_sharing': ['0', "Enable memory sharing"],
    'monitor': ['0', "Device model to use monitor"],
    'netif': ['0', "Make the domain a network interface backend."],
    'nomigrate': ['0', "Disable migration capability"],
    'oos': ['1', "Enable out-of-sync shadow page tabled"],
    'opengl': ['1', "Enable Open Graphics Library"],
    'pae': ['1', "Enable Physical Address Extension."],
    'pci_msitranslate': ['1', "Enable global PCI MSI-INTx translation"],
    'pci_power_mgmt': ['0', "Enable global PCI Power Management"],
    'sdl': ['0', "Device model to use Simple DirectMedia Layer (SDL)"],
    'superpages': ['0', "Create domain with 2 Meg superpages"],
    'suppress_spurious_page_faults': ['0', "Suppress injecting spurious page faults into guest"],
    'timer_mode': ['1', "Set Timer Mode to wallclock or delay virtual time"],
    'tpmif': ['0', "Make the domain a Trusted Platform Module interface backend."],
    'usb': ['0', "Emulate Universal Serial Bus devices"],
    'viridian': ['0', "Expose the Viridian interface"],
    'virtio': ['0', "Enable KVM Virtualization IO"],
    'vnc': ['1', "Device model to use VNC"],
    'vpt_align': ['1', "Enable aligning all Virtual Periodic Timers to reduce timer interrupts."],
    'xen_platform_pci': ['1', "Use Xen platform PCI"],
}

advanced_kvm_args = {
    'acpi': ['1', "Enable Advanced Configuration and Power Interface."],
    'apic': ['0', "Enable Advanced Programmable Interrupt Controller."],
    'hpet': ['1', "Enable virtual High-Precision Event Timer."],
    'pae': ['1', "Enable Physical Address Extension."],
    'usb': ['0', "Emulate Universal Serial Bus devices"],
    'virtio': ['0', "Enable KVM Virtualization IO"],
}

# Parse command line data
def parse_string(options, adv_opt):
    if adv_opt == "":
        return
    params = adv_opt.split(',')
    for p in params:
        opt, value = p.split('=')
        if not opt or not value:
            raise ParamError(err=ParamError.E_INVALID, details=p)
        if not value.isdigit() or int(value) != 0 and int(value) != 1:
            raise ParamError(err=ParamError.E_INVALID, details=p)
        (ok,state) = set_advanced_state(opt, int(value), options)
        if not ok:
            raise ParamError(err=ParamError.E_INVALID, details=p)

def _get_advanced_args():
    adv_args = None
    if caps.is_kvm() or caps.is_qemu():
        adv_args = advanced_kvm_args
    else:
        adv_args = advanced_xen_args
    return adv_args

# Returns a list object
def get_advanced_args():
    args_list = []
    adv_args = _get_advanced_args()
    for arg in adv_args:
        args_list.append(arg)
    args_list.sort()
    return args_list

# Load the list with the options and set the checkbox state
def init_advanced_settings_list(list, defaults):
    i = 0
    adv_args = _get_advanced_args()
    for arg in adv_args:
        # usb is used internally as a flag but has no libvirt xml type
        if arg == 'usb' and (caps.is_kvm() or caps.is_qemu()):
            continue
        if arg == 'virtio' and caps.is_xen():
            continue
        config_option = arg
        description = adv_args[arg][1]
        list.append([False, config_option, description])
        iter = list.get_iter(i,)
        i = i + 1
        current_state = 0
        if arg == 'acpi':
            current_state = defaults.acpi if (defaults.acpi is not None) else int(adv_args[arg][0])
        elif arg == 'apic':
            current_state = defaults.apic if (defaults.apic is not None) else int(adv_args[arg][0])
        elif arg == 'blkif':
            current_state = defaults.blkif if (defaults.blkif is not None) else int(adv_args[arg][0])
        elif arg == 'hap':
            current_state = defaults.hap if (defaults.hap is not None) else int(adv_args[arg][0])
        elif arg == 'hpet':
            current_state = defaults.hpet if (defaults.hpet is not None) else int(adv_args[arg][0])
        elif arg == 'isa':
            current_state = defaults.isa if (defaults.isa is not None) else int(adv_args[arg][0])
        elif arg == 'localtime':
            current_state = defaults.localtime if (defaults.localtime is not None) else int(adv_args[arg][0])
        elif arg == 'memory_sharing':
            current_state = defaults.memory_sharing if (defaults.memory_sharing is not None) else int(adv_args[arg][0])
        elif arg == 'monitor':
            current_state = defaults.monitor if (defaults.monitor is not None) else int(adv_args[arg][0])
        elif arg == 'netif':
            current_state = defaults.netif if (defaults.netif is not None) else int(adv_args[arg][0])
        elif arg == 'nomigrate':
            current_state = defaults.monitor if (defaults.monitor is not None) else int(adv_args[arg][0])
        elif arg == 'oos':
            current_state = defaults.oos if (defaults.oos is not None) else int(adv_args[arg][0])
        elif arg == 'opengl':
            current_state = defaults.opengl if (defaults.opengl is not None) else int(adv_args[arg][0])
        elif arg == 'pae':
            current_state = defaults.pae if (defaults.pae is not None) else int(adv_args[arg][0])
        elif arg == 'pci_msitranslate':
            current_state = defaults.pci_msitranslate if (defaults.pci_msitranslate is not None) else int(adv_args[arg][0])
        elif arg == 'pci_power_mgmt':
            current_state = defaults.pci_power_mgmt if (defaults.pci_power_mgmt is not None) else int(adv_args[arg][0])
        elif arg == 'sdl':
            current_state = defaults.sdl if (defaults.sdl is not None) else int(adv_args[arg][0])
        elif arg == 'superpages':
            current_state = defaults.superpages if (defaults.superpages is not None) else int(adv_args[arg][0])
        elif arg == 'suppress_spurious_page_faults':
            current_state = defaults.suppress_spurious_page_faults if (defaults.suppress_spurious_page_faults is not None) else int(adv_args[arg][0])
        elif arg == 'timer_mode':
            current_state = defaults.timer_mode if (defaults.timer_mode is not None) else int(adv_args[arg][0])
        elif arg == 'tpmif':
            current_state = defaults.tpmif if (defaults.tpmif is not None) else int(adv_args[arg][0])
        elif arg == 'usb':
            current_state = defaults.usb if (defaults.usb is not None) else int(adv_args[arg][0])
        elif arg == 'viridian':
            current_state = defaults.viridian if (defaults.viridian is not None) else int(adv_args[arg][0])
        elif arg == 'virtio':
            current_state = defaults.virtio if (defaults.virtio is not None) else int(adv_args[arg][0])
        elif arg == 'vnc':
            current_state = defaults.vnc if (defaults.vnc is not None) else int(adv_args[arg][0])
        elif arg == 'vpt_align':
            current_state = defaults.vpt_align if (defaults.vpt_align is not None) else int(adv_args[arg][0])
        elif arg == 'xen_platform_pci':
            current_state = defaults.xen_platform_pci if (defaults.xen_platform_pci is not None) else int(adv_args[arg][0])
        list.set_value(iter, 0, current_state)

# Set a specific option based on the state
def set_advanced_state(arg, state, options, vmdefaults=None):
    if arg == 'acpi':
        options.acpi = state
    elif arg == 'apic':
        if vmdefaults:
            state = vmdefaults.apic()
        options.apic = state
    elif arg == 'blkif':
        options.blkif = state
    elif arg == 'hap':
        options.hap = state
    elif arg == 'hpet':
        if vmdefaults:
            state = vmdefaults.hpet()
        options.hpet = state
    elif arg == 'isa':
        options.isa = state
    elif arg == 'localtime':
        options.localtime = state
    elif arg == 'memory_sharing':
        options.memory_sharing = state
    elif arg == 'netif':
        options.netif = state
    elif arg == 'monitor':
        options.monitor = state
    elif arg == 'nomigrate':
        options.nomigrate = state
    elif arg == 'oos':
        options.oos = state
    elif arg == 'opengl':
        options.opengl = state
    elif arg == 'pae':
        options.pae = state
    elif arg == 'pci_msitranslate':
        options.pci_msitranslate = state
    elif arg == 'pci_power_mgmt':
        options.pci_power_mgmt = state
    elif arg == 'sdl':
        options.sdl = state
    elif arg == 'superpages':
        options.superpages = state
    elif arg == 'suppress_spurious_page_faults':
        options.suppress_spurious_page_faults = state
    elif arg == 'timer_mode':
        options.timer_mode = state
    elif arg == 'tpmif':
        options.tpmif = state
    elif arg == 'usb':
        if vmdefaults:
            state = vmdefaults.usb()
        options.usb = state
    elif arg == 'viridian':
        options.viridian = state
    elif arg == 'virtio':
        if vmdefaults:
            state = vmdefaults.virtio()
        options.virtio = state
    elif arg == 'vnc':
        options.vnc = state
    elif arg == 'vpt_align':
        options.vpt_align = state
    elif arg == 'xen_platform_pci':
        options.xen_platform_pci = state
    else:
        return (False, state)
    return (True, state)

# Set the state of an option based on what was checked in the list
def set_advanced_arg(list, item_offset, state, options):
    arg = list[item_offset][1]
    arg = arg.split(' ')[0]
    set_advanced_state(arg, state, options)

# Reset to defaults
def reset_advanced_options_to_defaults(list, options, vmdefaults):
    items = _get_advanced_args().items()
    items.sort()
    i = 0
    for arg in items:
        # usb is used internally as a flag but has no libvirt xml type
        if arg[0] == 'usb' and (caps.is_kvm() or caps.is_qemu()):
            continue
        if arg[0] == 'virtio' and caps.is_xen():
            continue
        state = int(arg[1][0])
        (ok, state) = set_advanced_state(arg[0], state, options, vmdefaults)
        iter = list.get_iter(i,)
        i = i + 1
        list.set_value(iter, 0, state)

# Return a string object containing all the options to be writen out to the
# configuration file.  Don't bother writing an option in the default state.
def get_advanced_features_xen(options):
    value = ""
    adv_args = _get_advanced_args()
    for arg in adv_args:
        if arg == 'acpi' and options.acpi is not None and \
            int(options.acpi) != int(adv_args[arg][0]):
            value += 'acpi=%d\n' % int(options.acpi)
        elif arg == 'apic' and options.apic is not None and \
            int(options.apic) != int(adv_args[arg][0]):
            value += 'apic=%d\n' % int(options.apic)
        elif arg == 'blkif' and options.blkif is not None and \
            int(options.blkif) != int(adv_args[arg][0]):
            value += 'blkif=%d\n' % int(options.blkif)
        elif arg == 'hap' and options.hap is not None and \
            int(options.hap) != int(adv_args[arg][0]):
            value += 'hap=%d\n' % int(options.hap)
        elif arg == 'hpet' and options.hpet is not None and \
            int(options.hpet) != int(adv_args[arg][0]):
            value += 'hpet=%d\n' % int(options.hpet)
        elif arg == 'isa' and options.isa is not None and \
            int(options.isa) != int(adv_args[arg][0]):
            value += 'isa=%d\n' % int(options.isa)
        elif arg == 'localtime' and options.localtime is not None and \
            int(options.localtime) != int(adv_args[arg][0]):
            value += 'localtime=%d\n' % int(options.localtime)
        elif arg == 'memory_sharing' and options.memory_sharing is not None and \
            int(options.memory_sharing) != int(adv_args[arg][0]):
            value += 'memory_sharing=%d\n' % int(options.memory_sharing)
        elif arg == 'monitor' and options.monitor is not None and \
            int(options.monitor) != int(adv_args[arg][0]):
            value += 'monitor=%d\n' % int(options.monitor)
        elif arg == 'netif' and options.netif is not None and \
            int(options.netif) != int(adv_args[arg][0]):
            value += 'netif=%d\n' % int(options.netif)
        elif arg == 'nomigrate' and options.nomigrate is not None and \
            int(options.nomigrate) != int(adv_args[arg][0]):
            value += 'nomigrate=%d\n' % int(options.nomigrate)
        elif arg == 'oos' and options.oos is not None and \
            int(options.oos) != int(adv_args[arg][0]):
            value += 'oos=%d\n' % int(options.oos)
        elif arg == 'opengl' and options.opengl is not None and \
            int(options.opengl) != int(adv_args[arg][0]):
            value += 'opengl=%d\n' % int(options.opengl)
        elif arg == 'pae' and options.pae is not None and \
            int(options.pae) != int(adv_args[arg][0]):
            value += 'pae=%d\n' % int(options.pae)
        elif arg == 'pci_msitranslate' and options.pci_msitranslate is not None and \
            int(options.pci_msitranslate) != int(adv_args[arg][0]):
            value += 'pci_msitranslate=%d\n' % int(options.pci_msitranslate)
        elif arg == 'pci_power_mgmt' and options.pci_power_mgmt is not None and \
            int(options.pci_power_mgmt) != int(adv_args[arg][0]):
            value += 'pci_power_mgmt=%d\n' % int(options.pci_power_mgmt)
        elif arg == 'sdl' and options.sdl is not None and \
            int(options.sdl) != int(adv_args[arg][0]):
            value += 'sdl=%d\n' % int(options.sdl)
        elif arg == 'superpages' and options.superpages is not None and \
            int(options.superpages) != int(adv_args[arg][0]):
            value += 'superpages=%d\n' % int(options.superpages)
        elif arg == 'suppress_spurious_page_faults' and options.suppress_spurious_page_faults is not None and \
            int(options.suppress_spurious_page_faults) != int(adv_args[arg][0]):
            value += 'suppress_spurious_page_faults=%d\n' % int(options.suppress_spurious_page_faults)
        elif arg == 'timer_mode' and options.timer_mode is not None and \
            int(options.timer_mode) != int(adv_args[arg][0]):
            value += 'timer_mode=%d\n' % int(options.timer_mode)
        elif arg == 'tpmif' and options.tpmif is not None and \
            int(options.tpmif) != int(adv_args[arg][0]):
            value += 'tpmif=%d\n' % int(options.tpmif)
        elif arg == 'usb' and options.usb is not None and \
            int(options.usb) != int(adv_args[arg][0]):
            value += 'usb=%d\n' % int(options.usb)
        elif arg == 'viridian':
            virt_support = VMDefaults[options.os_type].virt_support()
            if virt_support[0] == 'hvm-ext':
                if options.viridian is None:
                    value += 'viridian=1\n'
                else:
                    value += 'viridian=%d\n' % int(options.viridian)
            elif options.viridian is not None and \
                int(options.viridian) != int(adv_args[arg][0]):
                value += 'viridian=%d\n' % int(options.viridian)
            else:
                value += 'viridian=0\n'
        elif arg == 'vnc' and options.vnc is not None and \
            int(options.vnc) != int(adv_args[arg][0]):
            value += 'vnc=%d\n' % int(options.vnc)
        elif arg == 'vpt_align' and options.vpt_align is not None and \
            int(options.vpt_align) != int(adv_args[arg][0]):
            value += 'vpt_align=%d\n' % int(options.vpt_align)
        elif arg == 'xen_platform_pci' and options.xen_platform_pci is not None and \
            int(options.xen_platform_pci) != int(adv_args[arg][0]):
            value += 'xen_platform_pci=%d\n' % int(options.xen_platform_pci)
    return value 

def get_advanced_features_kvm(options):
    value = ""
    adv_args = _get_advanced_args()
    for arg in adv_args:
        new_value = None
        if arg == 'acpi':
            if (options.acpi is None and adv_args[arg][0] == '1') or options.acpi is True:
                new_value = '<acpi/>'
        elif arg == 'apic':
            if (options.apic is None and adv_args[arg][0] == '1') or options.apic is True:
                new_value = '<apic/>'
        elif arg == 'pae':
            if (options.pae is None and adv_args[arg][0] == '1') or options.pae is True:
                new_value = '<pae/>'
        if new_value:
            if len(value):
                value += '\n    '
            else:
                value += '  '
            value += new_value
    return value 
