# Probes capabilities of the platform
#
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import os
import platform

from exceptions import *

def cpu_flags():
    f = open('/proc/cpuinfo')
    lines = f.readlines()
    f.close()
    for line in lines:
        if not line.startswith('flags'):
            continue
        flags = line[:-1].split(':', 1)[1]
        flst = flags.split(' ')
        return flst
    return []

def cpu_processors():
    f = open('/proc/cpuinfo')
    lines = f.readlines()
    f.close()
    num_processors = 0
    for line in lines:
        if not line.startswith('processor'):
            continue
        num_processors += 1
    if num_processors == 0:
        num_processors = 1
    return num_processors

def vcpu_processors(max_vcpus):
    cpus = cpu_processors()
    if ( cpus > max_vcpus ):
        cpus = max_vcpus
    return cpus

def free_memory():
    f = open('/proc/meminfo')
    lines = f.readlines()
    f.close()
    free_memory = 0
    for line in lines:
        if line.startswith('MemFree:') or line.startswith('Inactive:'):
            mem = line.split(':', 1)[1]
            mem = mem.strip()
            mem = mem.split(' ')[0]
            free_memory = free_memory + ((int)(mem) / 1024)
    return free_memory

_arch = None
def arch():
    global _arch
    if not _arch:
        a = os.uname()[4]
        if a in ('i386', 'i586', 'i686'):
            _arch = 'i586'
        else:
            _arch = a
    return _arch

def compatible_arches(supports_32nonpae=True):
    if is_ia64_capable():
        arches = [['ia64', 'xen']]
    elif is_x86_64_capable():
        if not supports_32nonpae:
            # SLE11 and newer 32bit support is pae only but is just 'kernel-xen', not 'kernel-xenpae'.
            arches = [['i386', 'xen'], ['i586', 'xen'], ['x86_64', 'xen']]
        else:
            arches = [['i386', 'xenpae'], ['i586', 'xenpae'], ['x86_64', 'xen']]
        # x86-64 can do either, but prefer to match dom0.
        if arch() == 'x86_64':
            arches.reverse()
    else:
        if is_pae_capable():
            if not supports_32nonpae:
                # SLE11 and newer 32bit support is pae only but is just 'kernel-xen', not 'kernel-xenpae'.
                arches = [['i386', 'xen'], ['i586', 'xen']]
            else:
                arches = [['i386', 'xenpae'], ['i586', 'xenpae']]
        else:
            arches = [['i386', 'xen'], ['i586', 'xen']]
    return arches

def kvm_arch():
    if not is_kvm():
        log.debug("kvm_arch.__del__")
    if platform.machine() == 'x86_64':
        return 'x86_64'
    return 'i686'

def xen_arch():
    if not is_xen():
        log.debug("xen_arch.__del__")
        raise XenError(XenError.E_NO_XEN)
    if is_ia64_capable():
        return 'ia64'
    if is_x86_64_capable():
        return 'x86_64'
    return 'i586'
    
_caps = []
def xen_caps():
    global _caps
    if not _caps:
        if os.path.exists('/sys/hypervisor/properties/capabilities'):
            _caps = open('/sys/hypervisor/properties/capabilities').read().split()
    return _caps

def is_ia64_capable():
    return 'xen-3.0-ia64' in xen_caps()

def is_x86_64_capable():
    return 'xen-3.0-x86_64' in xen_caps()

def is_pae_capable():
    return 'xen-3.0-x86_32p' in xen_caps()

def is_hvm_capable():
    if is_kvm() or is_qemu():
        return True
    caps = xen_caps()
    return ('hvm-3.0-x86_32' in caps or
            'hvm-3.0-x86_32p' in caps or
            'hvm-3.0-x86_64' in caps)

def is_blktap_capable():
    return os.path.exists('/dev/xen/blktap0')

def is_kvm():
    return os.path.exists('/dev/kvm')

def is_xen():
    return os.path.isdir('/proc/xen')

def is_qemu():
    if is_xen():
        return False
    return os.path.exists('/usr/bin/qemu')

def is_dom0():
    # FIXME
    return os.path.isdir('/proc/xen')

def is_domU():
    return is_xen() and not is_dom0()

def get_os_type():
    # Could check SPIdent on sles
    if os.path.exists('/etc/SuSE-release'):
        f = open('/etc/SuSE-release')
        lines = f.readlines()
        f.close()
        for line in lines:
            if line.startswith('openSUSE 10'):
                return 'opensuse'
            if line.startswith('openSUSE 11'):
                return 'opensuse11'
            if line.startswith('SUSE Linux Enterprise Server 10'):
                return 'sles10'
            if line.startswith('SUSE Linux Enterprise Desktop 10'):
                return 'sled10'
            if line.startswith('SUSE Linux Enterprise Server 11'):
                return 'sles11'
            if line.startswith('SUSE Linux Enterprise Desktop 11'):
                return 'sled11'
    return 'sles11'
