# Fully virtualized guest support
#
# Initially derived from virt-inst, Copyright 2005-2006 Red Hat, Inc.
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

from exceptions import *
from full_virt_guest import FullVirtGuest
from xen_guest import XenGuest
from vmdisks import Disk
from log import log
import vminstall.advanced

class FullVirtXenGuest(FullVirtGuest, XenGuest):
    def __init__(self, options):
        XenGuest.__init__(self, options)

        self._cdrom = None
        self.features = {
            "acpi": True,
            "pae": True,  # required to install 64 bit!
        }
        self._qemu = '/usr/lib/xen/bin/qemu-dm'

    def _get_features_xml(self):
        ret = []
        if self.options.apic:
            ret.append('<apic/>')
        for (k, v) in self.features.items():
            if v:
                ret.append('<%s/>' % k)
        return ret

    def _get_features_xen(self):
        ret = vminstall.advanced.get_advanced_features_xen(self.options)
        for (k, v) in self.features.items():
            if v:
                ret += '%s=1\n' %(k,)
        return ret

    def _get_mouse_xen(self):
        ret = ''
        if self.options.usb:
            #ret += "usb=1\n"
            if self.options.os_type != 'sles9':
                if self.options.usb_mouse_abs:
                    ret += "usbdevice='tablet'\n"
                else:
                    ret += "usbdevice='mouse'\n"
        return ret
        
    def GetConfigXml(self):
        if self.isInstall:
            if type(self.options.source) is Disk:
                if self.options.source.vdevtype == Disk.DEVICE_DISK:
                    bootdev = 'hd'
                else:
                    bootdev = 'cdrom'
            else:
                bootdev = 'cdrom'
        else:
            bootdev = 'hd'

        return [ "<domain type='xen'>",
                   self.getCommonXml(),
                 [ '<os>',
                   [ '<type>hvm</type>',
                     '<loader>%s</loader>' % self.options.loader,
                     "<boot dev='%s'/>" % bootdev ],
                   '</os>',
                   '<features>',
                     self._get_features_xml(),
                   '</features>',
                   '<devices>',
                   [ '<emulator>%s</emulator>' % self._qemu ],
                     self._get_disk_xml(),
                     self._get_network_xml(),
                     self._get_graphics_xml(),
                   '</devices>' ],
                 '</domain>' ]

    def GetConfigXen(self):
        if self.isInstall:
            if self.options.pxe_boot:
                bootdev = 'n'
            else:
                if type(self.options.source) is Disk:
                    if self.options.source.vdevtype == Disk.DEVICE_DISK:
                        bootdev = 'c'
                    else:
                        bootdev = 'd'
                else:
                    bootdev = 'd'
        else:
            bootdev = 'c'

        return """%(common)s
builder="hvm"
device_model="%(qemu)s"
kernel="%(loader)s"
boot="%(bootdev)s"
%(disks)s
%(networks)s
%(cpuid)s
%(cpucheck)s
%(graphics)s
%(features)s
%(mouse)s
serial="pty"
""" % { 'common': self.getCommonXen(), 'disks': self._get_disk_xen(),
        'networks': self._get_network_xen(), 'cpuid': self._get_cpuid_xen(), 'cpucheck': self._get_cpu_check_xen(), 
        'qemu': self._qemu, 'graphics': self._get_graphics_xen(), 'features': self._get_features_xen(),
        'bootdev': bootdev, 'loader': self.options.loader, 'mouse': self._get_mouse_xen()}

    def PostInstallCheck(self):
        """Verify that the first hard disk is bootable."""
        hds = filter(lambda disk: disk.vdevtype == Disk.DEVICE_DISK, self.options.disks)
        if hds:
            vdev = hds[0].block_attach()
            try:
                if hds[0].has_bootsector():
                    return
            finally:
                hds[0].block_detach()
        raise VmError(VmError.E_NO_BOOTLOADER)

