# GTK-based interface for vm-install
#
# Initially derived from virt-manager, Copyright 2005-2006 Red Hat, Inc.
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import gtk
import gobject
import gtk.gdk
import gtk.glade
import os
import string
import sys
import threading
import time

from disk_widgets import *
from gjob import GJob
from nic_widgets import *
from vminstall.job import Job
from vminstall.options import *
from vminstall.VMDefaults import VMDefaults as VMDefaults
from vminstall.globals import *
import vminstall
import vminstall.libvirt_console
import vminstall.xen_console
import vminstall.nics
import vminstall.options
import vminstall.hypervisor
import vminstall.advanced


def strify(name):
    if not name:
        return ''
    return str(name)

def set_combo_from_list(combo, items):
    """Setup a ComboBox or ComboBoxEntry based on a list of strings."""
    model = gtk.ListStore(gobject.TYPE_STRING)
    for item in items:
        model.append([item])
    combo.set_model(model)
    if type(combo) == gtk.ComboBoxEntry:
        if combo.get_text_column() == -1:
            combo.set_text_column(0)
    elif type(combo) == gtk.ComboBox:
        combo.clear()
        cell = gtk.CellRendererText()
        combo.pack_start(cell, True)
        combo.add_attribute(cell, 'text', 0)

#class NotebookPage:
#    def __init__(self, key, title):
#        self.key = key
#        self.root = gtk.VBox(homogeneous=False)
#        align = gtk.Alignment(xalign=0.0, yalign=0.5, xscale=0.0, yscale=1.0)
#        align.set_padding(5, 5, 6, 6)
#        label = gtk.Label()
#        label.set_single_line_mode(True)
#        label.set_markup('<span weight="heavy" size="xx-large" foreground="#FFF" background="#000">%s</span>' % title)
#        label.set_padding(5, 6)
#        align.add(label)
#        event = gtk.EventBox()
#        event.add(align)
#        event.modify_bg(gtk.STATE_NORMAL, gtk.gdk.color_parse('#000'))
#        self.root.pack_start(event, expand=False)
#        align = gtk.Alignment(xalign=0.0, yalign=0.0, xscale=1.0, yscale=1.0)
#        align.set_padding(0, 0, 20, 0)
#        self.content = align
#        self.root.pack_start(self.content, expand=True)
#        self.root.show_all()

################################################################################
# Class: VMCreate
#
# Class used to control the GUI interface.
#
################################################################################
class VMCreate(gobject.GObject):
    __gsignals__ = {
        'action-show-console': (gobject.SIGNAL_RUN_FIRST,
                                gobject.TYPE_NONE, (str,str)),
        'action-show-terminal': (gobject.SIGNAL_RUN_FIRST,
                                gobject.TYPE_NONE, (str,str)),
        'vmmcreate-closing': (gobject.SIGNAL_RUN_FIRST,
                              gobject.TYPE_NONE, (int,)),
    }

    def refresh_hypervisor_info(self):
        try:
            self.hypervisor_info = vminstall.hypervisor.connection.info()
        except:
            # An error will be displayed later, once the GUI is up.
            pass

    def __init__(self, job=None, options=None, defaults=None, virtman=False, key=0):
        self.__gobject_init__()

        if os.getuid() != 0:
            raise Error(Exception(), vminstall.msg.must_be_root)

        # Specific to HVM guests
        self.has_balloon_driver = ['opensuse11','sled11','sles11']

        if caps.is_xen():
            paths.load_xen_paths()
        elif caps.is_kvm():
            vminstall.paths.load_kvm_paths()
        elif caps.is_qemu():
            vminstall.paths.load_qemu_paths()
        else:
            raise HypervisorError(HypervisorError.E_NO_HYPERVISOR)

        self.job = job or Job(background=False)
        if options is None:
            options = vminstall.options.Options()
            defaults = vminstall.options.Options()
            vminstall.options.calculate(options, defaults)
        self.options = options
        self.defaults = defaults
        self.original_os_type = self.options.os_type

        self.hypervisor_info = None
        self.refresh_hypervisor_info()

        self.virtman = virtman
        self.key = key

        self.current_disk_index = None
        self.disk_caller_page = None
        self.nics_view = None
        self.current_nic = None
        self.current_nic_index = None

        self.os_settings_folder = os.getcwd()

        self.WINDOW_NAME = 'vmm-create'
        self.window = gtk.glade.XML(vminstall.paths.gladedir + 'vm-install.glade', self.WINDOW_NAME)
        self.topwin = self.window.get_widget(self.WINDOW_NAME)
        self.topwin.set_title(vminstall.msg.title_intro)

        self.summary = gtk.glade.XML(vminstall.paths.gladedir + 'vm-install.glade', 'summary-vbox')
        self.summary_vbox = self.summary.get_widget('summary-vbox')

        self.disks_widget = DisksWidget(allow_reorder=True, remove_fn=self.do_remove_disk,
                                        edit_fn=self.do_edit_disk, add_cdrom_fn=self.do_new_cdrom,
                                        add_harddisk_fn=self.do_new_harddisk)
        self.window.get_widget('disks-vbox').pack_end(self.disks_widget.get_widget())
        self.disk_edit_widget = DiskEditWidget(VMDefaults[defaults.os_type].virtio())
        self.window.get_widget('disk-edit-vbox').pack_end(self.disk_edit_widget.get_widget())

        # Connect the button event with the method that handles the event.
        # This allows us to go from someone clicking on a headline
        # (e.g. "Virtualization Method", "Hardware", etc.) to the method
        # that handles that page.
        self.summary.signal_autoconnect({
            'on_headline_method_button_release_event': self.do_method,
            'on_headline_vm_properties_button_press_event': self.do_vm_properties,
            'on_headline_hardware_button_press_event': self.do_hardware,
            'on_headline_graphics_button_release_event': self.do_graphics,
            'on_headline_disks_button_press_event': self.do_disks,
            'on_headline_network_button_press_event': self.do_network,
            'on_headline_os_button_press_event': self.do_os,
            })

        self.window.signal_autoconnect({
            'on_create_pages_switch_page': self.page_changed,
            'on_create_cancel_clicked': self.cancel,
            'on_vmm_create_delete_event': self.close,
            'on_create_back_clicked': self.backward,
            'on_create_forward_clicked': self.forward,
            'on_create_apply_clicked': self.apply,
            'on_create_ok_clicked': self.okay,
            'on_create_upgrade_clicked': self.upgrade,
            'on_pv_src_disk_add_clicked': self.do_new_pv_cdrom,
            'on_fv_src_disk_add_clicked': self.do_new_fv_cdrom,
            'on_button_nic_up_clicked': self.do_nic_up,
            'on_button_nic_down_clicked': self.do_nic_down,
            'on_button_new_nic_clicked': self.do_new_nic,
            'on_button_edit_nic_clicked': self.do_edit_nic,
            'on_button_delete_nic_clicked': self.do_delete_nic,
            'on_fv_install_source_focus_in': self.activate_fv_pxe_src,
            'on_pv_install_source_focus_in': self.activate_pv_network_src,
            'on_memory_startup_value_changed': self.set_max_memory,
            'on_create_memory_max_value_changed': self.set_max_memory,
            'on_fv_media_toggled': self.change_fv_media_type,
            'on_pv_media_toggled': self.change_pv_media_type,
            'on_os_settings_browse_clicked': self.browse_os_settings,
            'on_os_type_list_button_release_event': self.select_os_type,
            'on_pxe_install_clicked' : self.change_pv_media_type,
            'on_os_install_toggled': self.generate_os_list_on_toggle,
            'on_config_file_toggled': self.config_file_toggled,
            'on_upgrade_find_config_clicked': self.upgrade_find_config_clicked,
            'on_advanced_settings_clicked': self.do_advanced_settings,
            'on_advanced_reset_to_defaults_clicked': self.do_advanced_reset_to_defaults,
            })

        self.set_initial_state()

    def generate_os_list(self):
        os_type_upgrade = None
        default = default_upgrade = None
        groups = {}
        model = gtk.TreeStore(str, str)
        for os_type in VMDefaults.keys():
            if self.defaults.install == OS_INSTALL_UPGRADE:
                if not VMDefaults[os_type].is_upgradeable():
                    continue
                if not os_type_upgrade:
                    os_type_upgrade = os_type
            group_name = eval('vminstall.msg.os_%s' % VMDefaults[os_type].group())
            if not groups.has_key(group_name):
                groups[group_name] = []
            groups[group_name].append(os_type)
        group_names = vminstall.util.sort_numeric_aware(groups.keys())
        for group_name in group_names:
            piter = model.append(None, [group_name, ''])
            oss = map(lambda os: (VMDefaults[os].name(), os), groups[group_name])
            oss = vminstall.util.sort_tuples_by_nth(oss, 0)
            for os in oss:
                model.append(piter, os)
                if self.defaults.os_type == os[1]:
                    default = (group_names.index(group_name), oss.index(os))
                elif os_type_upgrade == os[1] and not default_upgrade:
                    default_upgrade = (group_names.index(group_name), oss.index(os))
            if default is None and default_upgrade is not None:
                default = default_upgrade
        treeview = self.window.get_widget('os-type-list')
        treeview.set_model(model)
        column = gtk.TreeViewColumn('os name', gtk.CellRendererText(), text=0)
        column.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
        treeview.append_column(column)
        treeview.set_search_column(0)
        treeview.set_expander_column(column)
        if default is not None:
            treeview.collapse_all()
            treeview.expand_row(default[:-1], False)
            treeselection = treeview.get_selection()
            treeselection.select_path(default)
        treeview.columns_autosize()
        treeview.show_all()

    def set_initial_state(self):
        self.notebook = self.window.get_widget('create-pages')
        self.notebook.set_show_tabs(False)
        self.pages = {}
        pages = self.notebook.get_n_pages()
        for i in xrange(pages):
            child = self.notebook.get_nth_page(i)
            label = self.notebook.get_tab_label_text(child)
            self.pages[label] = i

#        page = NotebookPage('test', "test")
#        self.notebook.insert_page(page.root, gtk.Label(page.key), 0)
#        page.content.add(gtk.Label("test"))
#        page.root.show_all()

        prompt = vminstall.msg.prompt
        bold_prompt = vminstall.msg.bold_prompt
        small = vminstall.msg.small

        # Set translated labels.  Not doing this within glade, so it's easier
        # to keep CLI & GUI in sync.
        black = gtk.gdk.color_parse('#000')
        for title in ('intro', 'install', 'os_type', 'summary', 'method',
                      'name', 'hardware', 'graphics', 'disks', 'disk',
                      'nics', 'nic_edit', 'fv_install', 'pv_install', 'pv_runtime',
                      'upgrade', 'advanced'):
            widget = self.window.get_widget('title_%s' % title)
            widget.modify_bg(gtk.STATE_NORMAL, black)
            label = widget.get_child()
            label.set_markup('<span weight="heavy" size="xx-large" foreground="#FFF">%s</span>' % eval('vminstall.msg.title_%s' % title))
        # 'os' is set later, since it can be 'os_install' or 'os_runtime'
        for name in ('method', 'name', 'hardware', 'graphics', 'disks', 'nics'):
            title = self.summary.get_widget('summary-title-%s' % name)
            text = eval('vminstall.msg.title_%s' % name)
            self._set_summary_title(title, text)
        for name in ('memory', 'max_memory', 'vcpus', 'operating_system',
                     'installation_source', 'automated_installation', 'extra_args'):
            widget = self.summary.get_widget('%s_prompt' % name)
            text = eval('vminstall.msg.%s_prompt' % name)
            widget.set_markup(bold_prompt(text))

        #
        # The next section uses:
        #       set_text
        #       set_radio_text
        # to create a bunch of widgets.
        #
        def set_text(name, widget=None, text=None, modifier=None, markup=False):
            if widget is None:
                widget = self.window.get_widget(name)
            vminstall.msg.set_text(name, widget, text, modifier, markup)
        def set_radio_text(name, text=None, modifier=None, markup=False):
            widget = self.window.get_widget(name).get_children()[0]
            vminstall.msg.set_text(name, widget, text, modifier, markup)
        set_radio_text('graphics_hw_cirrus')
        set_radio_text('graphics_hw_none')
        set_radio_text('graphics_hw_para')
        set_radio_text('graphics_hw_vesa')
        set_radio_text('nic-mac-random', text=vminstall.msg.random_mac)
        set_radio_text('nic-mac-specified', text=vminstall.msg.specified_mac)
        set_radio_text('os_install_no')
        set_radio_text('os_install_yes')
        if caps.is_kvm() or caps.is_qemu():
            self.window.get_widget('os_install_upgrade').hide_all()
            #self.window.get_widget('advanced-settings-button').hide_all()
        else:
            set_radio_text('os_install_upgrade')
        set_radio_text('pv-src-disk', text=vminstall.msg.virtual_disk, modifier=prompt)
        set_radio_text('pv-src-network', text=vminstall.msg.network_url, modifier=prompt)
        set_radio_text('pv-src-pxe', text=vminstall.msg.PXE_boot)
        set_radio_text('fv-src-disk', text=vminstall.msg.virtual_disk, modifier=prompt)
        set_radio_text('fv-src-pxe', text=vminstall.msg.PXE_boot, modifier=prompt)
        set_radio_text('virt-method-fv', text=vminstall.msg.full_virtualization)
        set_radio_text('virt-method-pv', text=vminstall.msg.paravirtualization)
        set_text('available_cpus_prompt', modifier=prompt)
        set_text('available_memory_prompt', modifier=prompt)
        set_text('fv_install_help', text=vminstall.msg.install_media_help)
        set_text('graphics_hw_help')
        set_text('graphics_note_none', modifier=small, markup=True)
        set_text('graphics_note_para', modifier=small, markup=True)
        set_text('keyboard_help', text=vminstall.msg.keyboard_hw_help)
        set_text('keymap-label', text=vminstall.msg.keymap, modifier=prompt)
        set_text('hardware_help')
        set_text('hardware-memory-title', text=vminstall.msg.memory, modifier=bold_prompt, markup=True)
        set_text('hardware-processors-note', text=vminstall.msg.processors_note, modifier=small, markup=True)
        set_text('hardware-processors-title', text=vminstall.msg.processors, modifier=bold_prompt, markup=True)
        set_text('i_extra_args_help', text=vminstall.msg.extra_args_help)
        set_text('i_extra_args_label', text=vminstall.msg.extra_args_prompt, modifier=prompt)
        set_text('pv_exit_args_help', text=vminstall.msg.exit_args_help)
        set_text('pv_on_poweroff_label', text=vminstall.msg.on_poweroff, modifier=prompt)
        set_text('pv_on_reboot_label', text=vminstall.msg.on_reboot, modifier=prompt)
        set_text('pv_on_crash_label', text=vminstall.msg.on_crash, modifier=prompt)
        set_text('pv_exit_args_help', text=vminstall.msg.exit_args_help)
        set_text('fv_exit_args_help', text=vminstall.msg.exit_args_help)
        set_text('fv_on_poweroff_label', text=vminstall.msg.on_poweroff, modifier=prompt)
        set_text('fv_on_reboot_label', text=vminstall.msg.on_reboot, modifier=prompt)
        set_text('fv_on_crash_label', text=vminstall.msg.on_crash, modifier=prompt)
        set_text('install_help')
        set_text('intro_help1')
        set_text('intro_help2', markup=True)
        set_text('intro_help3', markup=True)
        set_text('intro_help4', markup=True)
        set_text('intro_help5', markup=True)
        set_text('intro_help6', text=vminstall.msg.intro_help6 % {'url': '<span underline="low">http://www.novell.com/documentation/technology/vm_server</span>'}, markup=True)
        set_text('max_memory_prompt', modifier=prompt)
        set_text('memory_prompt', modifier=prompt)
        set_text('method_help')
        set_text('method_no_full_bios', text=vminstall.msg.exc_arch_novt_bios, modifier=small, markup=True)
        set_text('method_no_full', text=vminstall.msg.exc_arch_novt, modifier=small, markup=True)
        set_text('method_no_full_os', text=vminstall.msg.exc_param_nofull, modifier=small, markup=True)
        set_text('method_no_para_os', text=vminstall.msg.exc_param_nopara, modifier=small, markup=True)
        set_text('name_help')
        set_text('name_name', text=vminstall.msg.name, modifier=prompt)
        set_text('vm_description', text=vminstall.msg.description, modifier=prompt)
        set_text('nic_edit_help')
        set_text('nic-mac-prompt', text=vminstall.msg.mac_address, modifier=prompt)
        if caps.is_xen():
            set_text('nic-mac-label', text="00:16:3e:")
        else:
            set_text('nic-mac-label', text="52:54:00:")
        set_text('nic-name-prompt', text=vminstall.msg.name, modifier=prompt)
        set_text('nic-source-prompt', text=vminstall.msg.source, modifier=prompt)
        set_text('nic-type-prompt', text=vminstall.msg.type, modifier=prompt)
        set_text('os_settings_help')
        set_text('os_settings_label', text=vminstall.msg.os_settings_prompt)
        set_text('os_type_help')
        set_text('pv_install_help', text=vminstall.msg.install_media_help)
        set_text('r_extra_args_help', text=vminstall.msg.extra_args_help)
        set_text('r_extra_args_label', text=vminstall.msg.extra_args_prompt, modifier=prompt)
        set_text('r_exit_args_help', text=vminstall.msg.exit_args_help)
        set_text('r_on_poweroff_label', text=vminstall.msg.on_poweroff, modifier=prompt)
        set_text('r_on_reboot_label', text=vminstall.msg.on_reboot, modifier=prompt)
        set_text('r_on_crash_label', text=vminstall.msg.on_crash, modifier=prompt)
        set_text('summary_help', markup=True)
        set_text('vcpus_prompt', modifier=prompt)
        set_text('os_upgrade_type', text=vminstall.msg.os_upgrade_type, modifier=prompt)
        widget = self.window.get_widget('os_upgrade_type').set_markup(bold_prompt(vminstall.msg.os_upgrade_type))
        set_text('virtual_machines', text=vminstall.msg.virtual_machines, modifier=prompt)
        set_radio_text('config-file', text=vminstall.msg.config_file)
        set_text('os_upgrade', text=vminstall.msg.upgrade)
        set_text('upgrade_network_url_help', text=vminstall.msg.upgrade_network_url_help)
        set_text('upgrade_network_url', text=vminstall.msg.network_url, modifier=prompt)
        set_text('advanced_help', text=vminstall.msg.advanced_help)
        set_text('advanced_settings', text=vminstall.msg.advanced_settings)
        self.window.get_widget('advanced-settings-button').set_label(vminstall.msg.advanced_settings)
        self.window.get_widget('reset-to-defaults').set_label(vminstall.msg.advanced_reset)

        # If we don't have full-virt support disable the choice, and
        # display a message telling the user why it is not working
        if vminstall.caps.is_hvm_capable():
            self.window.get_widget("virt-method-fv").set_sensitive(True)
            self.window.get_widget("virt-method-fv-unsupported").hide()
            self.window.get_widget("virt-method-fv-disabled").hide()
        else:
            self.window.get_widget("virt-method-fv").set_sensitive(False)
            flags = vminstall.caps.cpu_flags()
            if "vmx" in flags or "svm" in flags:
                self.window.get_widget("virt-method-fv-unsupported").hide()
                self.window.get_widget("virt-method-fv-disabled").show()
            else:
                self.window.get_widget("virt-method-fv-unsupported").show()
                self.window.get_widget("virt-method-fv-disabled").hide()
        #self.window.get_widget("pxe_combo").hide()

        # Add OSs to os_type notebook page
        self.generate_os_list()

        # Add NICs summary
        self.nics_view = NicsView([])
        self.window.get_widget('nics-scrolled-window').get_child().add(self.nics_view.view)
        self.nics_view.view.show_all()

        # Add summary
        scroll = self.window.get_widget('scrolled-summary')
        scroll.add_with_viewport(self.summary_vbox)

        self._reset_state()

    def _reset_state(self):
        self.notebook.set_current_page(self.pages['intro'])
        # Notebook is not yet displayed, so signal doesn't get emitted, so
        # load page defaults manually.
        self.load_page_defaults(self.pages['intro'])
        self.change_pv_media_type()

    def reset_state(self):
        self._reset_state()
        self.options = vminstall.options.Options()
        self.defaults = vminstall.options.Options()

    def show(self):
        self.topwin.show()

        hand = gtk.gdk.Cursor(gtk.gdk.HAND2)
        for name in ('method', 'name', 'hardware', 'graphics', 'disks', 'nics', 'os'):
            widget = self.summary.get_widget('summary-title-%s-eb' % name)
            widget.realize()
            widget.window.set_cursor(hand)

        self.topwin.present()

    def is_visible(self):
        if self.topwin.flags() & gtk.VISIBLE:
            return True
        return False

    def _set_summary_title(self, widget, text):
        widget.set_markup('<big><span foreground="blue">%s</span></big>' % text)

    def graphics_str(self):
        g = self.defaults.graphics
        text = eval('vminstall.msg.graphics_hw_' + g)
        return text

    def disks_str(self):
        text = ''
        if self.defaults.disks:
            for i in xrange(len(self.defaults.disks)):
                text += '%d: %s\n' % (i+1, self.defaults.disks[i])
        if not text:
            text = vminstall.msg.none
        return text.strip()

    def nics_str(self):
        text = ''
        if self.defaults.nics:
            for i in xrange(len(self.defaults.nics)):
                text += '%d: %s\n' % (i+1, str(self.defaults.nics[i]))
        if not text:
            text = vminstall.msg.none
        return text.strip()

    def select_os_type(self, widget, event):
        if event.type == gtk.gdk._2BUTTON_PRESS:
            self.forward()

    def do_new_pv_cdrom(self, button):
        self.disk_caller_page = 'pv_install'
        self.disks_widget.set_disks(self.defaults.disks)
        self.do_new_cdrom(button)

    def do_new_fv_cdrom(self, button):
        self.disk_caller_page = 'fv_install'
        self.disks_widget.set_disks(self.defaults.disks)
        self.do_new_cdrom(button)

    def do_new_cdrom(self, button):
        disk = vmdisks.Disk(pdev=self.disk_edit_widget.default_cdrom(),
                            vdevType=vmdisks.Disk.DEVICE_CDROM,
                            full_virt=self.defaults.full_virt,
                            virtio=self.defaults.virtio,
                            disk_group=self.disks_widget.get_disks())
        self._do_new_disk(disk)

    def do_new_harddisk(self, button):
        disk = vmdisks.Disk(pdev=None, vdevType=vmdisks.Disk.DEVICE_DISK,
                            full_virt=self.defaults.full_virt,
                            virtio=self.defaults.virtio,
                            disk_group=self.disks_widget.get_disks())
        self._do_new_disk(disk)

    def _do_new_disk(self, disk):
        disk.update(vmname=self.defaults.name)
        self.disk_edit_widget.set(disk)
        self.notebook.set_current_page(self.pages['disk'])
        return True

    def do_remove_disk(self, disk):
        remaining_disks = self.disks_widget.get_disks()
        for disk in remaining_disks:
            disk.disk_group = remaining_disks

    def do_edit_disk(self, button):
        (disk, index) = self.disks_widget.get_selected()
        if disk is None:
            return False
        self.current_disk_index = index
        self.disk_edit_widget.set(disk.copy())

        log.debug("Editing disk '%s'" % str(disk))
        self.notebook.set_current_page(self.pages['disk'])
        return True

    def do_pxe_combo(self):
        combo=self.window.get_widget('pv-src-pxe-box')
        combo.show()
        saved_os_type = self.options.os_type
        saved_source = self.options.source
        self.original_os_type = self.options.os_type
        self.options.os_type='pxe'
        self.options.source='pxe'
        action=VMDefaults[self.options.os_type].actions()
        try:
            kernel=action.getPXEconfigFromPXEserver(self.options)
        except Exception, e:
            self._exc_box(e)
            self.options.os_type=saved_os_type
            self.options.source=saved_source
            self.original_os_type = None
            return
        src = [str(label) for label in kernel]
        set_combo_from_list(combo, src)

    def config_file_toggled(self, widget=None):
        if not widget:
            widget = self.window.get_widget('config-file')
        text_widget = self.window.get_widget('config-fullname')
        #scrolled_widget = self.window.get_widget('virtual-machine-scrolled-window')
        if widget.get_active():
            text_widget.set_sensitive(True)
            #scrolled_widget.set_sensitive(False)
        else:
            text_widget.set_sensitive(False)
            #scrolled_widget.set_sensitive(True)
     
    def upgrade_find_config_clicked(self, src, ignore=None):
        self.window.get_widget('config-file').set_active(1)
        file=None
        folder = self.window.get_widget('config-fullname').get_text()
        if not os.path.exists(folder):
            path = os.path.dirname(folder)
            if os.path.exists(path):
                folder = path
            else:
                folder = '/etc/xen/vm/'
        if not os.path.isdir(folder):
            file = os.path.basename(folder)
            folder = os.path.dirname(folder)

        (file, folder) = chooser(vminstall.msg.locate_os_settings,
                                 self.topwin, file=file,
                                 mode=gtk.FILE_CHOOSER_ACTION_OPEN,
                                 folder=folder)
        if file != None:
            self.os_settings_folder = folder
            self.window.get_widget('config-fullname').set_text(file)
 
    def do_advanced_settings(self, widget=None):
        self.notebook.set_current_page(self.pages['advanced'])
        return True

    def do_advanced_reset_to_defaults(self, widget=None):
        sw = self.window.get_widget('advanced-scrolled-window')
        view = sw.get_child()
        list = view.get_model()
        vminstall.advanced.reset_advanced_options_to_defaults(list, self.options, VMDefaults[self.defaults.os_type])

    #---------------------------------------------------------------------------
    # Class: VMCreate
    # Method: load_page_defaults
    #
    # Parms:
    #   page_number:    The page number that we are to load the defaults.
    #
    # With the given page_number, load up the page defaults for each
    # of the elements within the page.
    #
    #---------------------------------------------------------------------------
    def load_page_defaults(self, page_number):
        """Initializes a page, loading the settings from 'defaults'."""
        self.set_page_buttons(page_number)

        #
        # kvm only works fully-virtualized; therefore, if
        # we running on a kvm hypervisor, set the default to 
        # be fully-virtualized.
        #
        if caps.is_kvm() or caps.is_qemu():
            self.window.get_widget('virt-method-fv').set_active(True)
            self.options.full_virt = True

        #
        # Page: Install an Operating System
        #
        if page_number == self.pages['install']:
            if self.defaults.install:
                if self.defaults.install == OS_INSTALL_UPGRADE:
                    self.window.get_widget('os_install_upgrade').set_active(True)
                else:
                    self.window.get_widget('os_install_yes').set_active(True)
            else:
                self.window.get_widget('os_install_no').set_active(True)
        #
        # Page: Type of Operating System
        #
        elif page_number == self.pages['os_type']:
            treeview = self.window.get_widget('os-type-list')
            treeselection = treeview.get_selection()
            rows = treeselection.get_selected_rows()[1]
            if len(rows):
                rows = rows[0]
                treeview.scroll_to_cell(rows)
        #
        # Page: Summary
        #
        elif page_number == self.pages['summary']:
            if self.defaults.full_virt:
                text = vminstall.msg.fully_virtualized
            else:
                text = vminstall.msg.paravirtualized

            # If we are running on kvm, disable the ablility to select
            # between fully-virtual and para-virtual by hiding the 
            # "Virtualization Method" option.
            if caps.is_kvm() or caps.is_qemu():
                self.summary.get_widget("summary-title-method-eb").hide()
                self.summary.get_widget("summary-title-method").hide()
                self.summary.get_widget("summary-method").hide()

            self.summary.get_widget("summary-method").set_text(text)
            self.summary.get_widget("summary-name").set_text(strify(self.defaults.name))
            self.summary.get_widget("summary-description").set_text(strify(self.defaults.description))
            self.summary.get_widget("summary-initial-memory").set_text(str(self.defaults.memoryMB) + " MB")
            if self.defaults.full_virt and \
                self.defaults.max_memoryMB > self.defaults.memoryMB and \
                self.defaults.os_type not in self.has_balloon_driver:
                self.summary.get_widget("summary-max-memory").set_markup('<span foreground="red">%s MB</span>' % self.defaults.max_memoryMB)
            else:
                self.summary.get_widget("summary-max-memory").set_text(str(self.defaults.max_memoryMB) + " MB")
            self.summary.get_widget("summary-virtual-cpus").set_text(str(int(self.defaults.vcpus)))
            self.summary.get_widget("summary-graphics").set_text(self.graphics_str())
            defined_disks = self.disks_str()
            if self.defaults.disks and not self.defaults.disks[0].validate():
                self.summary.get_widget("summary-disks").set_markup('<span foreground="red">%s</span>' % defined_disks)
            else:
                self.summary.get_widget("summary-disks").set_text(defined_disks)
            self.summary.get_widget("summary-nics").set_text(self.nics_str())

            title = self.summary.get_widget('summary-title-os')
            self.summary.get_widget("summary-os-type").set_text(VMDefaults[self.defaults.os_type].name())
            src_row = ('installation_source_prompt', 'summary-install-source')
            auto_row = ('automated_installation_prompt', 'summary-os-settings')
            args_row = ('extra_args_prompt', 'summary-extra-args')
            def show_row(row):
                for name in row:
                    self.summary.get_widget(name).show()
            def hide_row(row):
                for name in row:
                    self.summary.get_widget(name).hide()
            if self.defaults.install:
                if self.defaults.source is None and not self.defaults.pxe_boot:
                    if not self.defaults.full_virt and self.defaults.os_type == default_os_type:
                        dom0_inst_source = util.getDom0InstallSource()
                        if dom0_inst_source:
                            self.defaults.source = dom0_inst_source
                            self.options.source = dom0_inst_source
                            vminstall.options.calculate(self.options, self.defaults)
                self._set_summary_title(title, vminstall.msg.title_os_install)
                src = strify(self.defaults.source)
                if self.defaults.pxe_boot:
                    src = vminstall.msg.PXE_boot
                if self.options.source:
                    if not self.defaults.pxe_boot and not self.defaults.full_virt:
                        src = strify(self.options.source)
                self.summary.get_widget("summary-install-source").set_text(src)
                show_row(src_row)
                if self.defaults.full_virt:
                    hide_row(auto_row)
                else:
                    text = ''
                    if self.defaults.os_settings:
                        auto_inst_type = VMDefaults[self.defaults.os_type].auto_install()
                        if auto_inst_type is None:
                            text = vminstall.msg.exc_param_noauto
                        else:
                            text = '%s, using %s' % (auto_inst_type, self.defaults.os_settings)
                    self.summary.get_widget("summary-os-settings").set_text(text)
                    show_row(auto_row)
            else:
                self._set_summary_title(title, vminstall.msg.title_os_runtime)
                hide_row(src_row)
                hide_row(auto_row)
            if self.defaults.full_virt:
                hide_row(args_row)
            else:
                self.summary.get_widget("summary-extra-args").set_text(self.defaults.extra_args)
                show_row(args_row)
        #
        # Page: Virtualization Method page.
        #
        elif page_number == self.pages['method']:
            for method in (('virt-method-pv-unsupported-by-os', 'virt-method-pv', self.defaults.para_possible),
                           ('virt-method-fv-unsupported-by-os', 'virt-method-fv', self.defaults.full_possible)):
                warn_widget = self.window.get_widget(method[0])
                widget = self.window.get_widget(method[1])
                if method[2]():
                    widget.set_sensitive(True)
                    warn_widget.hide()
                else:
                    widget.set_sensitive(False)
                    warn_widget.show()
            if self.defaults.full_virt:
                name = 'virt-method-fv'
            else:
                name = 'virt-method-pv'
            self.window.get_widget(name).set_active(True)
        #
        # Page: Name of the Virtual Machine
        #
        elif page_number == self.pages['name']:
            if self.defaults.description:
                self.window.get_widget('description-entry').set_text(self.defaults.description)
            w = self.window.get_widget('create-vm-name')
            w.connect('activate', self.apply)
            w.set_text(strify(self.defaults.name))
            w.grab_focus()
        #
        # Page: Hardware
        #
        elif page_number == self.pages['hardware']:
            self.refresh_hypervisor_info()
            self.window.get_widget("create-cpus-physical").set_text(str(self.hypervisor_info['nr_cpus']))
            self.set_max_memory()
            startup = self.window.get_widget("memory-startup").get_adjustment()
            self.window.get_widget("host-memory-free").set_text('%d MB' % startup.upper)
            self.window.get_widget("memory-max").set_value(self.defaults.max_memoryMB)
            self.window.get_widget("memory-startup").set_value(self.defaults.memoryMB)
            self.window.get_widget("create-vcpus").set_value(self.defaults.vcpus)
        #
        # Page: Graphics
        #
        # If fully-virtual:
        #   cirrus: active
        #   vesa: active
        #   para: inactive
        #   para-help: inactive
        #
        # If para-virtual:
        #   cirrus: inactive
        #   vesa: inactive
        #   para: active
        #   para-help: active
        #
        # Then set the default graphics adapter to active.
        #
        elif page_number == self.pages['graphics']:
            fv = self.defaults.full_virt
            for name in ('graphics_hw_cirrus', 'graphics_hw_vesa'):
                self.window.get_widget(name).set_sensitive(fv)
            for name in ('graphics_hw_para', 'graphics_hw_para_help'):
                self.window.get_widget(name).set_sensitive(not fv)
            self.window.get_widget('graphics_hw_' + self.defaults.graphics).set_active(True)
            combo = self.window.get_widget('keymap-combobox')
            kmaps = vminstall.keytable.read_keymaps()
            set_combo_from_list(combo, kmaps)
            if self.options.keymap:
                index = kmaps.index(self.options.keymap)
            else:
                keymap = keytable.get_keymap()
                if keymap:
                    index = kmaps.index(keymap)
                else:
                    index = 0
            combo.set_active(index)
        # Page: Disks
        #
        elif page_number == self.pages['disks']:
            disks = self.disks_widget.get_disks()
            for d in disks:
                d.disk_group = disks
            max_disks = len(vmdisks.get_possible_vdevs(self.defaults.full_virt, self.defaults.virtio))
            self.disks_widget.set_max_disks(max_disks)
        #
        # Page: Disk
        #
        elif page_number == self.pages['disk']:
            self.disk_edit_widget.refresh()
        #
        # Page: Network Adapters
        #
        elif page_number == self.pages['nics']:
            self.window.get_widget('nics-scrolled-window').get_child().grab_focus()
            self.current_nic = self.current_nic_index = None
            num_nics = self.nics_view.rows()
            for button in ('button-nic-up', 'button-nic-down'):
                self.window.get_widget(button).set_sensitive(num_nics > 1)
            for button in ('button-edit-nic', 'button-delete-nic'):
                self.window.get_widget(button).set_sensitive(num_nics > 0)
            for button in ('button-new-nic', ):
                self.window.get_widget(button).set_sensitive(num_nics < vminstall.hypervisor.max_nics)
        #
        # Page: NIC edit
        #
        elif page_number == self.pages['nic_edit']:
            nic = self.current_nic
            assert(nic is not None)
            self.window.get_widget('nic-name').set_text(str(nic.number))

            combo = self.window.get_widget('nic-type')
            if caps.is_kvm() or caps.is_qemu():
                nic_models = vminstall.util.sort_dict_keys_by_value(vminstall.nics.KvmNic.models)
                set_combo_from_list(combo, [vminstall.nics.KvmNic.models[m] for m in nic_models])
                combo.set_active(nic_models.index(nic.model))
            else:
                if not self.defaults.full_virt:
                    para_nic = []
                    para_nic.append(vminstall.nics.XenNic.models.get('para'))
                    set_combo_from_list(combo, para_nic)
                    combo.set_active(0)
                else:
                    nic_models = vminstall.util.sort_dict_keys_by_value(vminstall.nics.XenNic.models)
                    set_combo_from_list(combo, [vminstall.nics.XenNic.models[m] for m in nic_models])
                    combo.set_active(nic_models.index(nic.model))

            bridges = [vminstall.msg.default]
            bridges.extend(vminstall.nics.get_bridges())
            combo = self.window.get_widget('nic-source')
            set_combo_from_list(combo, bridges)
            if nic.bridge in bridges:
                combo.set_active(bridges.index(nic.bridge))
            else:
                combo.set_active(0)

            if nic.mac is None:
                self.window.get_widget('nic-mac-random').set_active(True)
            else:
                self.window.get_widget('nic-mac-specified').set_active(True)
                bytes = nic.mac.split(':')
                assert(len(bytes) == 6)
                for i in (1, 2, 3):
                    self.window.get_widget('mac%d' % i).set_text(bytes[-i])
        #
        # Page: Fully virtual install
        #
        elif page_number == self.pages['fv_install']:
            combo = self.window.get_widget('fv-src-disk-box')
            self.src_disks = vmdisks.get_fv_boot_disks(self.defaults.disks)
            src_desc = [str(disk) for disk in self.src_disks]
            set_combo_from_list(combo, src_desc)
            source = self.defaults.source
            if len(self.src_disks):
                if type(source) is vmdisks.Disk and source in self.src_disks:
                    combo.set_active(self.src_disks.index(source))
                else:
                    combo.set_active(0)
            if self.options.pxe_boot:
                self.window.get_widget('fv-src-pxe').set_active(True)
            else:
                self.window.get_widget('fv-src-disk').set_active(True)
            on_exit_normal = on_exit[:]
            on_exit_normal.pop()
            on_exit_normal.pop()
            combo = self.window.get_widget('fv-poweroff-combo')
            set_combo_from_list(combo, on_exit_normal)
            index = on_exit_normal.index(self.defaults.on_poweroff)
            combo.set_active(index)
            combo = self.window.get_widget('fv-reboot-combo')
            set_combo_from_list(combo, on_exit_normal)
            index = on_exit_normal.index(self.defaults.on_reboot)
            combo.set_active(index)
            combo = self.window.get_widget('fv-crash-combo')
            set_combo_from_list(combo, on_exit)
            index = on_exit.index(self.defaults.on_crash)
            combo.set_active(index)
        #
        # Page: Paravirtual install
        #
        elif page_number == self.pages['pv_install']:
            combo = self.window.get_widget('pv-src-disk-box')
            self.src_disks = filter(lambda disk: disk.exists(), self.defaults.disks)
            src_desc = [str(disk) for disk in self.src_disks]
            set_combo_from_list(combo, src_desc)
            source = self.defaults.source
            if len(self.src_disks):
                if type(source) is vmdisks.Disk and source in self.src_disks:
                    combo.set_active(self.src_disks.index(source))
                else:
                    combo.set_active(0)
            if type(source) is vmdisks.Disk:
                self.window.get_widget('pv-src-disk').set_active(True)
            elif self.options.os_type == 'pxe':
                self.window.get_widget('pv-src-pxe').set_active(True)
            else:
                self.window.get_widget('pv-src-network').set_active(True)
                combo = self.window.get_widget('pv-src-network-url')
                (index, inst_sources) = util.getInstallSources()
                set_combo_from_list(combo, inst_sources)
                src = strify(source)
                if src:
                    if src not in inst_sources:
                        combo.insert_text(0, src)
                        combo.set_active(0)
                    else:
                        index = inst_sources.index(src)
                        combo.set_active(inst_sources.index(src))
                else:
                    if index != -1 and self.defaults.os_type == default_os_type:
                        combo.set_active(index)
                    else:
                        combo.set_active(0)
            autotype = VMDefaults[self.defaults.os_type].auto_install()
            sens = (autotype is not None)
            if not autotype:
                autotype = vminstall.msg.os_settings_prompt
            autotype += ':'
            self.window.get_widget('os_settings_label').set_text(autotype)
            self.window.get_widget('os-settings').set_text(self.defaults.os_settings)
            for l in ('os_settings_help', 'os-settings-hbox'):
                self.window.get_widget(l).set_sensitive(sens)
            self.window.get_widget('i-extra-args').set_text(self.defaults.extra_args)
            on_exit_normal = on_exit[:]
            on_exit_normal.pop()
            on_exit_normal.pop()
            combo = self.window.get_widget('pv-poweroff-combo')
            set_combo_from_list(combo, on_exit_normal)
            index = on_exit_normal.index(self.defaults.on_poweroff)
            combo.set_active(index)
            combo = self.window.get_widget('pv-reboot-combo')
            set_combo_from_list(combo, on_exit_normal)
            index = on_exit_normal.index(self.defaults.on_reboot)
            combo.set_active(index)
            combo = self.window.get_widget('pv-crash-combo')
            set_combo_from_list(combo, on_exit)
            index = on_exit.index(self.defaults.on_crash)
            combo.set_active(index)
        #
        # Page: Paravirtual Runtime
        #
        elif page_number == self.pages['pv_runtime']:
            w = self.window.get_widget('r-extra-args')
            w.connect('activate', self.apply)
            w.set_text(self.defaults.extra_args)
            w.grab_focus()
            on_exit_normal = on_exit[:]
            on_exit_normal.pop()
            on_exit_normal.pop()
            combo = self.window.get_widget('r-poweroff-combo')
            set_combo_from_list(combo, on_exit_normal)
            index = on_exit_normal.index(self.defaults.on_poweroff)
            combo.set_active(index)
            combo = self.window.get_widget('r-reboot-combo')
            set_combo_from_list(combo, on_exit_normal)
            index = on_exit_normal.index(self.defaults.on_reboot)
            combo.set_active(index)
            combo = self.window.get_widget('r-crash-combo')
            set_combo_from_list(combo, on_exit)
            index = on_exit.index(self.defaults.on_crash)
            combo.set_active(index)
        #
        # Page: Upgrade
        #
        elif page_number == self.pages['upgrade']:
            name = VMDefaults[self.defaults.os_type].name()
            self.window.get_widget('os-upgrade-name').set_text(name)
            # Populate the ScrolledList of managed VMs
            list = gtk.ListStore(gobject.TYPE_STRING)
            view = gtk.TreeView(list)
            column = gtk.TreeViewColumn(vminstall.msg.name, gtk.CellRendererText(), text=0)
            view.set_property("headers-visible", False)
            view.append_column(column)
            managed_vms = vminstall.util.get_running_vms()
            for vm in managed_vms:
                list.set_value(list.append(),0,vm)
            win = self.window.get_widget('virtual-machine-scrolled-window')
            child = win.get_child()
            old_view = child.get_child()
            if old_view is not None:
                child.remove(old_view)
            child.add(view)
            win.show_all()
            widget = self.window.get_widget('config-file')
            text_widget = self.window.get_widget('config-fullname')
            if self.defaults.upgrade_config:
                self.window.get_widget('config-fullname').set_text(self.defaults.upgrade_config)
            if len(managed_vms) == 0:
                widget.set_active(1)
            else:
                if widget.get_active():
                    text_widget.set_sensitive(True)
                else:
                    text_widget.set_sensitive(False)
            if self.defaults.source:
                self.window.get_widget('upgrade-network-url-text').set_text(self.defaults.source)
        #
        # Page: Advanced
        #
        elif page_number == self.pages['advanced']:
            # Populate the ScrolledList of supported config options

            def column_toggled(cell, path, user_data):
                model, options = user_data
                # Get the toggled iter and value from the column 0 checkbox
                iter = model.get_iter((int(path),))
                value = model.get_value(iter, 0)
                # Toggle the value
                value = not value
                # Set the new value
                model.set(iter, 0, value)
                vminstall.advanced.set_advanced_arg(model, path, value, self.options)

            # Create a ListStore with two columns
            list = gtk.ListStore(gobject.TYPE_BOOLEAN, str, str)
            # Fill in the list with the advanced options
            vminstall.advanced.init_advanced_settings_list(list, self.defaults)
            # Create the TreeView using our list
            view = gtk.TreeView(list)
            #view.set_property("headers-visible", False)
            view.expand_all()
            # Create a CellRendererText to render the data
            cr_option_text = gtk.CellRendererText()
            cr_desc_text = gtk.CellRendererText()
            cr_toggle = gtk.CellRendererToggle()
            cr_toggle.connect('toggled', column_toggled, (list, self.options))
            # Create the TreeViewColumns to display the data
            tvc_option_text = gtk.TreeViewColumn(vminstall.msg.option)
            tvc_desc_text = gtk.TreeViewColumn(vminstall.msg.description)
            tvc_toggle = gtk.TreeViewColumn(vminstall.msg.enabled, cr_toggle, active=0)
            # Add the TreeViewColumns to the TreeView
            view.append_column(tvc_toggle)
            view.append_column(tvc_option_text)
            view.append_column(tvc_desc_text)
            # Add the text cells to the tree view column and allow it to expand
            tvc_option_text.pack_start(cr_option_text, True)
            tvc_desc_text.pack_start(cr_desc_text, True)
            # Set the cell "text" attribute to column 1 - retrieve text
            # from that column in treestore
            tvc_option_text.add_attribute(cr_option_text, 'text', 1)
            tvc_desc_text.add_attribute(cr_desc_text, 'text', 2)
            # Make column 1 searchable
            view.set_search_column(1)
            # Allow sorting on column 1
            #tvc_option_text.set_sort_column_id(1)
            # Set sort order
            list.set_sort_column_id(1,gtk.SORT_ASCENDING)
            sw = self.window.get_widget('advanced-scrolled-window')
            old_view = sw.get_child()
            if old_view is not None:
                sw.remove(old_view)
            sw.add(view)
            sw.show_all()

    ##
    ##  Buttons
    ##

    def cancel(self, ignore=None):
        page_number = self.notebook.get_current_page()
        if page_number in self.intro_pages():
            self.close()
        elif page_number == self.pages['disk']:
            disk_caller_page = self.disk_caller_page or 'disks'
            self.notebook.set_current_page(self.pages[disk_caller_page])
            self.current_disk_index = None
        elif page_number == self.pages['nic_edit']:
            self.notebook.set_current_page(self.pages['nics'])
        elif page_number == self.pages['advanced']:
            self.notebook.set_current_page(self.pages['hardware'])
        else:
            self.notebook.set_current_page(self.pages['summary'])

    def okay(self, ignore=None):
        page_number = self.notebook.get_current_page()
        if page_number == self.pages['summary']:
            if type(self.options.source) is str:
                if self.options.source and self.options.source.startswith('smb://') and self.options.smboptions is None:
                    self.getsmbpasswd()
            self.finish()
        else:
            self.apply()

    def apply(self, ignore=None):
        if self.save_page(self.notebook.get_current_page()):
            page_number = self.notebook.get_current_page()
            if page_number == self.pages['advanced']:
                self.notebook.set_current_page(self.pages['hardware'])
            else:
                if page_number == self.pages['hardware'] and self.defaults.full_virt and \
                    self.defaults.max_memoryMB > self.defaults.memoryMB and \
                    self.defaults.os_type not in self.has_balloon_driver:
                    self._warn_box(vminstall.msg.warning, vminstall.msg.maximum_memory_warning)
                self.notebook.set_current_page(self.pages['summary'])

    def upgrade(self, ignore=None):
        # Validate a virtual machine was selected from one of the following,
        #   a. virtual-machine-scrolled-window
        #   b. config-fullname text box.
        log.debug("Upgrade an existing operating system")
        self.options.upgrade = True
        self.options.install = OS_INSTALL_UPGRADE
        self.defaults.install = OS_INSTALL_UPGRADE
        if self.window.get_widget('config-file').get_active():
            log.debug("Configuration file specified for upgrade.")
            config_file = self.window.get_widget('config-fullname').get_text()
            # Open config and lookup name field (config name and name field should match)
            vm_name = vminstall.util.get_config_file_option(config_file, "name")
            if not vm_name:
                vm_name = os.path.basename(config_file)
            self.options.upgrade_managed_vm = False
            self.options.upgrade_config = config_file
            if hypervisor.connection.domain(vm_name):
                # User is pointing at a config file, but the VM is also managed
                log.debug("Configuration file specified points to a Managed VM.")
                self.options.upgrade_managed_vm = True
        else:
            log.debug("Managed VM specified for upgrade.")
            viewp = self.window.get_widget('virtual-machine-scrolled-window').get_child()
            tree_selection = viewp.get_child().get_selection()
            (lstore, iter) = tree_selection.get_selected()
            if iter is None or lstore is None:
                return
            vm_name = lstore[iter][0]
            self.options.upgrade_managed_vm = True
        self.options.name = vm_name
        # Check to see if the VM is running
        if hypervisor.connection.domain_running(vm_name):
            log.debug("Validation failed: VM must not be running for an upgrade")
            self._exc_box(VmError(err=VmError.E_VM_RUNNING))
            return
        # Initialize the internal data from the config file. Do this before calling calculate
        try:
            vminstall.options.initialize_vm_from_config_file(self.options, vm_name)
        except Exception, e:
            log.debug("Failed to read VM's configuration data ")
            self._exc_box(e)
            self.load_page_defaults(self.pages['upgrade'])
            return
        # Validate the network install location found in upgrade-network-url-text widget
        self.options.source = self.window.get_widget('upgrade-network-url-text').get_text()
        if not self.options.source:
            log.debug("validation failed: No install source specified for upgrade")
            self._exc_box(InstSrcError(err=InstSrcError.E_UNDEF))
            return
        try:
            vminstall.options.calculate(self.options, self.defaults)
        except Exception, e:
            log.debug("Validation failed: Install source protocol or other parameter is invalid for upgrade")
            self._exc_box(e)
            return
        self.finish()

    def intro_pages(self):
        return [self.pages[name] for name in ['intro', 'install', 'os_type', 'summary', 'upgrade']]

    def valid_intro_pages(self):
        valid_pages = []
        if not self.options.os_settings or not self.original_os_type:
            valid_pages.append(self.pages['intro'])
        if not self.options.os_settings:
            valid_pages.append(self.pages['install'])
        if not self.original_os_type:
            valid_pages.append(self.pages['os_type'])
        if self.defaults.install == OS_INSTALL_UPGRADE:
            valid_pages.append(self.pages['upgrade'])
        else:
            valid_pages.append(self.pages['summary'])
        return valid_pages

    def valid_forward_pages(self, current_page):
        pages = self.intro_pages()
        forward_pages = pages[pages.index(current_page)+1:]
        valid_pages = self.valid_intro_pages()
        return filter(lambda page: page in valid_pages, forward_pages)

    def valid_backward_pages(self, current_page):
        pages = self.intro_pages()
        pages.reverse()
        backward_pages = pages[pages.index(current_page)+1:]
        valid_pages = self.valid_intro_pages()
        return filter(lambda page: page in valid_pages, backward_pages)

    def forward(self, ignore=None):
        if self.hypervisor_info is None:
            log.error("The hypervisor is not running")
            self._error_box(vminstall.msg.error, vminstall.msg.exc_xen_none)
            self.close()
            return
        current_page = self.notebook.get_current_page()
        if self.save_page(current_page):
            self.notebook.set_current_page(self.valid_forward_pages(current_page)[0])

    def backward(self, ignore=None):
        current_page = self.notebook.get_current_page()
        self.notebook.set_current_page(self.valid_backward_pages(current_page)[0])

    def close(self, ignore1=None, ignore2=None):
        log.info("vm-install #%d is closing..." % self.key)
        self.job.do_cancel(forceful=False)
        self.topwin.hide()
        self.topwin.destroy()
        self.emit('vmmcreate-closing', self.key)

    def set_page_buttons(self, page_number):
        if page_number in self.intro_pages():
            self.window.get_widget("create-apply").hide()
            if page_number == self.pages['upgrade']:
                self.window.get_widget("create-ok").hide()
                self.window.get_widget('create-forward').hide()
                self.window.get_widget("create-cancel").show()
                self.window.get_widget('create-back').show()
                self.window.get_widget("create-upgrade").show()
                self.window.get_widget('create-upgrade').set_sensitive(1)
            else:
                forward = (len(self.valid_forward_pages(page_number)) > 0)
                backward = (len(self.valid_backward_pages(page_number)) > 0)
                self.window.get_widget('create-back').show()
                self.window.get_widget('create-back').set_sensitive(backward)
                self.window.get_widget("create-cancel").show()
                if page_number == self.pages['summary']:
                    self.window.get_widget("create-upgrade").hide()
                    self.window.get_widget('create-ok').show()
                    self.window.get_widget('create-forward').hide()
                else:
                    self.window.get_widget("create-upgrade").hide()
                    self.window.get_widget('create-ok').hide()
                    self.window.get_widget('create-forward').show()
                    self.window.get_widget('create-forward').set_sensitive(forward)
            return
        if page_number == self.pages['disk']:
            self.window.get_widget("create-upgrade").hide()
            self.window.get_widget("create-forward").hide()
            self.window.get_widget("create-back").hide()
            self.window.get_widget("create-apply").hide()
            self.window.get_widget("create-ok").show()
            self.window.get_widget("create-cancel").show()
        else:
            self.window.get_widget("create-upgrade").hide()
            self.window.get_widget("create-forward").hide()
            self.window.get_widget("create-back").hide()
            self.window.get_widget("create-ok").hide()
            self.window.get_widget("create-cancel").show()
            self.window.get_widget("create-apply").show()
            if page_number == self.pages['hardware']:
                if caps.is_xen():
                    full_virt = self.options.full_virt if (self.options.full_virt is not None) else self.defaults.full_virt
                    if not full_virt:
                        self.window.get_widget('advanced-settings-button').hide_all()
                    else:
                        self.window.get_widget('advanced-settings-button').show_all()

    def do_method(self, widget, event):
        self.notebook.set_current_page(self.pages['method'])

    def do_vm_properties(self, widget, event):
        self.notebook.set_current_page(self.pages['name'])

    def do_hardware(self, widget, event):
        self.notebook.set_current_page(self.pages['hardware'])

    def do_graphics(self, widget, event):
        self.notebook.set_current_page(self.pages['graphics'])

    def do_disks(self, widget, event):
        self.disks_widget.set_disks(self.defaults.disks)
        self.notebook.set_current_page(self.pages['disks'])

    def do_nic_up(self, button):
        for row in self.nics_view.move_selected_up():
            nic = self.nics_view.get_row(row)
            self.nics_view.update(row, nic)
        return True

    def do_nic_down(self, button):
        for row in self.nics_view.move_selected_down():
            nic = self.nics_view.get_row(row)
            self.nics_view.update(row, nic)
        return True

    def do_new_nic(self, button):
        if caps.is_kvm() or caps.is_qemu():
            self.current_nic = vminstall.nics.KvmNic(number=self.nics_view.rows()+1,
                                                     full_virt=self.defaults.full_virt)
        else:
            self.current_nic = vminstall.nics.XenNic(number=self.nics_view.rows()+1,
                                                     full_virt=self.defaults.full_virt)
        self.notebook.set_current_page(self.pages['nic_edit'])
        return True

    def do_edit_nic(self, button):
        (nic, index) = self.nics_view.get_selected()
        if nic is None:
            return False
        self.current_nic = nic.copy()
        self.current_nic_index = index
        log.debug("Editing nic '%s'" % str(self.current_nic))
        self.notebook.set_current_page(self.pages['nic_edit'])
        return True

    def do_delete_nic(self, button):
        self.nics_view.delete_selected()
        self.load_page_defaults(self.pages['nics'])
        return True

    def do_network(self, widget, event):
        nics = self.defaults.nics
        log.debug("Re-loading nics table with '%s'" % map(str, nics))
        self.nics_view.set(nics)
        self.notebook.set_current_page(self.pages['nics'])

    def do_os(self, widget, event):
        if self.defaults.install:
            if self.defaults.install == OS_INSTALL_UPGRADE:
                self.notebook.set_current_page(self.pages['upgrade'])
                return
            if self.defaults.full_virt:
                self.notebook.set_current_page(self.pages['fv_install'])
            else:
                self.notebook.set_current_page(self.pages['pv_install'])
        else:
            if self.defaults.full_virt:
                pass  # No FV runtime configuration possible
            else:
                self.notebook.set_current_page(self.pages['pv_runtime'])

    def page_changed(self, notebook, page, page_number):
        self.load_page_defaults(page_number)

    def finish(self, ignore=None):
        if self.defaults.install != OS_INSTALL_UPGRADE:
            try:
                vminstall.options.finalize(self.defaults)
            except Exception, e:
                self._exc_box(e)
                return

        def async_ok(button):
            self.topwin.set_sensitive(True)
            config_win.destroy()

        # Create progress window
        def async_cancel(button):
            self.job.do_cancel(forceful=True)

        if self.defaults.no_install == True:
            # Start job to write config file
            errors = []
            gjob = GJob(self.job, self.defaults, errors)
            gjob.start()
            # Create config written window
            config_glade = gtk.glade.XML(vminstall.paths.gladedir + 'vm-install.glade', "vmm-config")
            config_glade.signal_autoconnect({ 'on_config_written_ok_clicked': async_ok, })
            config_glade.get_widget('job_conf_written').set_text(vminstall.msg.job_conf_written)
            config_win = config_glade.get_widget("vmm-config")
            config_win.set_title(vminstall.msg.config_title)
            self.topwin.set_sensitive(False)
            config_win.set_transient_for(self.topwin)
            config_win.present()
            return

        pbar_glade = gtk.glade.XML(vminstall.paths.gladedir + 'vm-install.glade', "vmm-progress")
        pbar_glade.signal_autoconnect({
            'on_progress_cancel_clicked': async_cancel,
        })
        pbar_glade.get_widget('please_wait').set_text(vminstall.msg.please_wait)
        pbar_win = pbar_glade.get_widget("vmm-progress")
        pbar_win.connect('delete-event', lambda w,e: True)
        pbar_win.set_title(vminstall.msg.progress_title)

        # Show progress window
        self.topwin.set_sensitive(False)
        pbar_win.set_transient_for(self.topwin)
        pbar_win.present()
        pbar_win.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.WATCH))

        # Create GObject Job wrapper
        pbar = pbar_glade.get_widget("pbar")
        label = pbar_glade.get_widget("status")
        errors = []

        # GTK 2.11 or PyGTK 2.10 seem to suffer from a threading bug.  Some
        # windowing functions are unsafe to call from another thread, even
        # while holding the GTK lock.  Use idle callbacks instead.  Novell bz
        # #306964
        def destroy_window(w):
            gtk.gdk.threads_enter()
            w.destroy()
            gtk.gdk.threads_leave()
        def hide_window(w):
            gtk.gdk.threads_enter()
            w.hide()
            gtk.gdk.threads_leave()
        def do_progress_change(src, text, fraction):
            gtk.gdk.threads_enter()
            log.debug("do_progress_change: %0.2f %s" % (fraction, text))
            if text:
                label.set_text(text)
            if fraction >= 0.0 and fraction <= 1.0:
                pbar.set_fraction(fraction)
            gtk.gdk.threads_leave()
        def do_progress_close(src):
            log.debug("do_progress_close")
            if pbar_win is None:
                return
            gtk.gdk.threads_enter()
            gobject.idle_add(destroy_window, pbar_win)
            self.topwin.set_sensitive(True)
            try:
                self.topwin.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.TOP_LEFT_ARROW))
            except:
                pass  # Parent's window may have closed
            gtk.gdk.threads_leave()
        def do_show_console(src, uuid):
            log.debug("do_show_console")
            if caps.is_kvm() or caps.is_qemu():
                guest_console = vminstall.libvirt_console.LibVirtConsole()
            else:
                guest_console = vminstall.xen_console.XenConsole()
            if self.virtman:
                domid = guest_console.get_domid_from_uuid(uuid)
                signal = 'action-show-terminal'
                if (self.defaults.graphics != 'none' and \
                    guest_console.get_vnc_port_from_domid(domid, timeout=10.0) != None):
                    signal = 'action-show-console'
                log.debug("%s of %s via virt-manager" % (signal, uuid))
                def poke_virt_man():
                    if not caps.is_xen():
                        self.emit(signal, 'qemu:///system', uuid)
                    else:
                        self.emit(signal, 'xen:///', uuid)
                    gobject.idle_add(hide_window, pbar_win)
                    gobject.idle_add(hide_window, self.topwin)
                    return False  # Do not call me again
                # virt-manager (and libvirt) poll.  Asking virt-manager to open
                # the console too soon after the VM has started will fail.
                gobject.timeout_add(1100, poke_virt_man)
                if self.job.state[0] == Job.INSTALL:
                    wait_on_me = uuid
                else:
                    wait_on_me = None
            else:
                wait_on_me = self.job._show_console()
            if type(wait_on_me) is str:
                guest_console.wait_for_uuid_to_exit(wait_on_me)
            elif type(wait_on_me) is int:
                gobject.idle_add(hide_window, pbar_win)
                gobject.idle_add(hide_window, self.topwin)
                (pid, status) = os.waitpid(wait_on_me, 0)
        def _do_show_error(sem):
            gtk.gdk.threads_enter()
            log.debug("_do_show_error")
            e = errors.pop()
            log.debug("_do_show_error: error was '%s'" % str(e))
            if type(e) is str:
                self._error_box(vminstall.msg.error, e)
            else:
                self._exc_box(e)
            sem.release()
            gtk.gdk.threads_leave()
        def do_show_error(src):
            log.debug("do_show_error")
            sem = threading.Semaphore()
            sem.acquire()
            gobject.idle_add(_do_show_error, sem)
            sem.acquire()
        def _do_done():
            gtk.gdk.threads_enter()
            log.debug("_do_done: state was '%s'" % self.job.state[2])
            if self.job.state[0] == self.job.FINISHED:
                self.close()
            else:
                self.topwin.present()
            gtk.gdk.threads_leave()
        def do_done(src):
            log.debug("do_done")
            gobject.idle_add(_do_done)
        log.info("Starting VM creation job (key=%d)..." % self.key)
        gjob = GJob(self.job, self.defaults, errors)
        gjob.connect('install-progress-change', do_progress_change)
        gjob.connect('install-progress-close', do_progress_close)
        gjob.connect('install-show-console', do_show_console)
        gjob.connect('install-show-error', do_show_error)
        gjob.connect('install-done', do_done)
        gjob.start()

    def activate_fv_pxe_src(self, ignore1=None, ignore2=None):
        self.window.get_widget('fv-src-pxe').activate()

    def activate_pv_network_src(self, ignore1=None, ignore2=None):
        self.window.get_widget('pv-src-network').activate()

    def browse_os_settings(self, src, ignore=None):
        file = self.window.get_widget('os-settings').get_text()
        (file, folder) = chooser(vminstall.msg.locate_os_settings,
                                 self.topwin, file=file,
                                 mode=gtk.FILE_CHOOSER_ACTION_OPEN,
                                 folder=self.os_settings_folder)
        if file != None:
            self.os_settings_folder = folder
            self.window.get_widget('os-settings').set_text(file)

    def change_fv_media_type(self, ignore=None):
        if self.window.get_widget("fv-src-disk").get_active():
            self.window.get_widget("fv-src-disk-box").set_sensitive(True)
            self.window.get_widget("fv-src-disk-add").set_sensitive(True)
        elif self.window.get_widget("fv-src-pxe").get_active():
            self.window.get_widget("fv-src-disk-box").set_sensitive(False)
            self.window.get_widget("fv-src-disk-add").set_sensitive(False)

    def change_pv_media_type(self, ignore=None):
        if self.window.get_widget("pv-src-disk").get_active():
            self.window.get_widget("pv-src-disk-box").set_sensitive(True)
            self.window.get_widget("pv-src-disk-add").set_sensitive(True)
            self.window.get_widget("pv-src-network-url").set_sensitive(False)
            self.window.get_widget("pv-src-pxe-box").set_sensitive(False)
            if self.original_os_type:
                self.options.os_type = self.original_os_type
                self.original_os_type = None
        elif self.window.get_widget("pv-src-network").get_active():
            self.window.get_widget("pv-src-disk-box").set_sensitive(False)
            self.window.get_widget("pv-src-disk-add").set_sensitive(False)
            self.window.get_widget("pv-src-network-url").set_sensitive(True)
            self.window.get_widget("pv-src-pxe-box").set_sensitive(False)
            if self.original_os_type:
                self.options.os_type = self.original_os_type
                self.original_os_type = None
        elif self.window.get_widget("pv-src-pxe").get_active():
            self.window.get_widget("pv-src-disk-box").set_sensitive(False)
            self.window.get_widget("pv-src-disk-add").set_sensitive(False)
            self.window.get_widget("pv-src-network-url").set_sensitive(False)
            self.window.get_widget("pv-src-pxe-box").set_sensitive(True)
            self.do_pxe_combo()

    def generate_os_list_on_toggle(self, ignore=None):
        initial_value = self.defaults.install
        if self.window.get_widget("os_install_yes").get_active():
            self.defaults.install = OS_INSTALL_YES 
        elif self.window.get_widget("os_install_no").get_active():
            self.defaults.install = OS_INSTALL_NO
        if self.window.get_widget("os_install_upgrade").get_active():
            self.defaults.install = OS_INSTALL_UPGRADE
        if initial_value != self.defaults.install:
            self.generate_os_list()

    def set_max_memory(self, src=None):
        if self.defaults.full_virt:
            avail_mb = self.hypervisor_info['max_hvm_memory']
        else:
            avail_mb = self.hypervisor_info['max_para_memory']
        maximum = self.window.get_widget("memory-max").get_adjustment()
        startup = self.window.get_widget("memory-startup").get_adjustment()
        init_mem, max_mem = vminstall.options.calculate_max_memory(self.options, startup.value, maximum.value)
        if self.options.full_virt:
            if startup.value > init_mem:
                startup.value = init_mem
        else:
            if startup.value > avail_mb:
                startup.value = avail_mb
        if startup.value > maximum.value:
            if maximum.value < max_mem:
                maximum.value = startup.value
            else:
                startup.value = maximum.value
        if maximum.value > max_mem:
            maximum.value = max_mem
        os_type_min = VMDefaults[self.options.os_type].min_memory()
        if startup.value < os_type_min and os_type_min <= avail_mb:
            startup.value = os_type_min
        startup.upper = init_mem
        maximum.upper = max_mem

    def save_page(self, page_num):
        """Returns True if ok to proceed to next page, else False.
        Transactional w.r.t options/defaults.
        """
        ok = False
        old_options = self.options.copy()
        try:
            ok = self._save_page(page_num)
        except Exception, e:
            self._exc_box(e)
            self.options = old_options
        if not ok:
            return False

        old_defaults = self.defaults.copy()
        try:
            vminstall.options.calculate(self.options, self.defaults)
        except Exception, e:
            self._exc_box(e)
            self.options = old_options
            self.defaults = old_defaults
            return False
        return True

    def _save_page(self, page_num):
        """Saves data from the specified page into 'options'.
        On sucess, returns True.
        Not transactional on failure, w.r.t. options/defaults.
        """
        if page_num == self.pages['intro']:
            pass
        elif page_num == self.pages['install']:
            self.options.install = self.window.get_widget('os_install_yes').get_active()
            if not self.options.install:
                if self.window.get_widget('os_install_upgrade').get_active():
                    self.options.install = OS_INSTALL_UPGRADE
        elif page_num == self.pages['os_type']:
            treeview = self.window.get_widget('os-type-list')
            selection = treeview.get_selection()
            (model, iter) = selection.get_selected()
            if iter is None:
                return False
            os_type = model.get_value(iter, 1)
            if not os_type:
                return False
            self.options.os_type = os_type
        elif page_num == self.pages['summary']:
            pass
        elif page_num == self.pages['hardware']:
            def get_spin_value(widget):
                widget.update()
                return widget.get_value()
            self.options.memoryMB = get_spin_value(self.window.get_widget("memory-startup"))
            self.options.max_memoryMB = get_spin_value(self.window.get_widget("memory-max"))
            self.options.vcpus = get_spin_value(self.window.get_widget("create-vcpus"))
        # Find the graphics device that's active, and set the 
        # options.graphics variable to that device.
        elif page_num == self.pages['graphics']:
            for hw in ('cirrus', 'none', 'para', 'vesa'):
                if self.window.get_widget('graphics_hw_' + hw).get_active():
                    self.options.graphics = hw
                    break
            combo = self.window.get_widget('keymap-combobox')
            self.options.keymap = combo.get_active_text()
        elif page_num == self.pages['method']:
            full_virt = not self.window.get_widget('virt-method-pv').get_active()
            full_virt_saved = self.options.full_virt
            self.options.full_virt = full_virt
            if not self.options.full_virt:
                self.options.pxe_boot = False
            elif full_virt_saved is not full_virt:
                self.options.os_settings = ''
                if self.options.source and type(self.options.source) == str:
                    isrc = self.options.source
                    if isrc.startswith('ftp://') or isrc.startswith('http://') or isrc.startswith('nfs://') or isrc.startswith('smb://'):
                        self.options.source =  None
        elif page_num == self.pages['name']:
            name = self.window.get_widget('create-vm-name').get_text()
            description = self.window.get_widget('description-entry').get_text()
            self.options.name = name
            self.options.description = description
        elif page_num == self.pages['disks']:
            self.options.disks = self.disks_widget.get_disks()
            for d in self.options.disks:
                d.disk_group = self.options.disks
        elif page_num == self.pages['disk']:
            disk = self.disk_edit_widget.get()

            # Verify that the new disk does not conflict.  If it does, the
            # transaction will roll back, and GUI disk list hasn't yet been changed.
            self.options.disks = self.disks_widget.get_disks()
            if self.current_disk_index is not None:
                self.options.disks[self.current_disk_index] = disk
            else:
                self.options.disks.append(disk)
            vminstall.options.calculate(self.options, self.defaults)

            # It does not conflict, so safe to add it to the disks widget:
            if self.current_disk_index is not None:
                self.disks_widget.set_index(self.current_disk_index, disk)
                self.current_disk_index = None
            else:
                self.disks_widget.append(disk)

            if self.disk_caller_page:
                self.options.source = disk
                ret_page = self.disk_caller_page
                self.disk_caller_page = None
            else:
                ret_page = 'disks'
            vminstall.options.calculate(self.options, self.defaults)
            self.notebook.set_current_page(self.pages[ret_page])
            return None
        elif page_num == self.pages['nics']:
            self.options.nics = self.nics_view.get()
            # They've looked at the NICs, so don't allow the default to change if full_virt changes.
            for nic in self.options.nics:
                nic.set_model(nic.get_model())
            self.nics_view.set([])
            log.debug("Retrieved nic list from nic table: '%s'" % map(str, self.options.nics))
        elif page_num == self.pages['nic_edit']:
            nic = self.current_nic
            assert(nic is not None)
            if caps.is_kvm() or caps.is_qemu():
                nic_models = vminstall.util.sort_dict_keys_by_value(vminstall.nics.KvmNic.models)
            else:
                nic_models = vminstall.util.sort_dict_keys_by_value(vminstall.nics.XenNic.models)
            if self.defaults.full_virt:
                nic.model = nic_models[self.window.get_widget('nic-type').get_active()]
            else:
                nic.model = 'para'

            if self.window.get_widget('nic-mac-random').get_active():
                nic.mac = None
            else:
                mac1 = self.window.get_widget('mac1').get_text()
                mac2 = self.window.get_widget('mac2').get_text()
                mac3 = self.window.get_widget('mac3').get_text()
                if caps.is_xen():
                    nic.mac = '00:16:3e:%s:%s:%s' % (mac3, mac2, mac1)
                else:
                    nic.mac = '52:54:00:%s:%s:%s' % (mac3, mac2, mac1)

            bridge = self.window.get_widget('nic-source').get_active()
            if bridge == 0:
                nic.bridge = None
            else:
                nic.bridge = vminstall.nics.get_bridges()[bridge-1]

            if self.current_nic_index is not None:
                self.nics_view.update(self.current_nic_index, self.current_nic)
            else:
                self.nics_view.append(self.current_nic)
            vminstall.options.calculate(self.options, self.defaults)
            self.notebook.set_current_page(self.pages['nics'])
            return None
        elif page_num == self.pages['fv_install']:
            if self.window.get_widget('fv-src-disk').get_active():
                i = self.window.get_widget('fv-src-disk-box').get_active()
                if i < 0:
                    raise vminstall.DiskResourceError(vminstall.DiskResourceError.E_NOT_FOUND)
                self.options.source = self.src_disks[i]
                self.options.pxe_boot = 0
            else:
                self.options.pxe_boot = 1
            combo = self.window.get_widget('fv-poweroff-combo')
            self.options.on_poweroff = combo.get_active_text()
            combo = self.window.get_widget('fv-reboot-combo')
            self.options.on_reboot = combo.get_active_text()
            combo = self.window.get_widget('fv-crash-combo')
            self.options.on_crash = combo.get_active_text()
        elif page_num == self.pages['pv_install']:
            if self.window.get_widget('pv-src-disk').get_active():
                i = self.window.get_widget('pv-src-disk-box').get_active()
                if i < 0:
                    raise vminstall.DiskResourceError(vminstall.DiskResourceError.E_NOT_FOUND)
                src = self.src_disks[i]
                self.options.pxe_boot = False
            elif self.window.get_widget("pv-src-network").get_active():
                src = self.window.get_widget('pv-src-network-url').get_active_text()
                self.options.pxe_boot = False
            else:
                src = None
                combo=self.window.get_widget('pv-src-pxe-box')
                if combo.get_active() != -1:
                    self.options.pxe_boot = True
                    labelN = combo.get_active_text()
                    action = VMDefaults[self.options.os_type].actions()
                    kernel = action.getKernelList()
                    src = labelN+' '+kernel[labelN]
                    self.options.pxe_label = labelN
                    self.defaults.pxe_label = labelN
            self.options.source = src
            self.options.os_settings = self.window.get_widget('os-settings').get_text()
            self.options.extra_args = self.window.get_widget('i-extra-args').get_text()
            combo = self.window.get_widget('pv-poweroff-combo')
            self.options.on_poweroff = combo.get_active_text()
            combo = self.window.get_widget('pv-reboot-combo')
            self.options.on_reboot = combo.get_active_text()
            combo = self.window.get_widget('pv-crash-combo')
            self.options.on_crash = combo.get_active_text()
        elif page_num == self.pages['pv_runtime']:
            self.options.extra_args = self.window.get_widget('r-extra-args').get_text()
            combo = self.window.get_widget('r-poweroff-combo')
            self.options.on_poweroff = combo.get_active_text()
            combo = self.window.get_widget('r-reboot-combo')
            self.options.on_reboot = combo.get_active_text()
            combo = self.window.get_widget('r-crash-combo')
            self.options.on_crash = combo.get_active_text()
        elif page_num == self.pages['upgrade']:
            self.options.source = self.window.get_widget('upgrade-network-url-text').get_text()
        elif page_num == self.pages['advanced']:
            vminstall.options.calculate(self.options, self.defaults)

        return True

    def _exc_box(self, exc):
        if isinstance(exc, vminstall.Error):
            self._error_box(vminstall.msg.error, text=exc.title, details=exc.details)
        else:
            self._error_box(vminstall.msg.error, text=str(exc))

    def _error_box(self, title, text=None, details=None):
        self.topwin.present()
        message_box = gtk.MessageDialog(self.topwin,
                                        0, gtk.MESSAGE_ERROR, gtk.BUTTONS_OK, title)
        if text:
            message_box.format_secondary_text(text)
        if details:
            expander = gtk.Expander("Details")
            tbuffer = gtk.TextBuffer()
            tbuffer.set_text(details)
            textview = gtk.TextView(tbuffer)
            expander.add(gtk.Label(details))
            expander.show_all()
            message_box.vbox.pack_end(expander)
        message_box.run()
        message_box.destroy()

    def _warn_box(self, title, text=None, details=None):
        self.topwin.present()
        message_box = gtk.MessageDialog(self.topwin,
                                        0, gtk.MESSAGE_WARNING, gtk.BUTTONS_OK, title)
        if text:
            message_box.format_secondary_text(text)
        if details:
            expander = gtk.Expander("Details")
            tbuffer = gtk.TextBuffer()
            tbuffer.set_text(details)
            textview = gtk.TextView(tbuffer)
            expander.add(gtk.Label(details))
            expander.show_all()
            message_box.vbox.pack_end(expander)
        message_box.run()
        message_box.destroy()

    def getsmbpasswd(self):
        dialog = gtk.Dialog(title="SMB PASSWORD DIALOG",
            parent=self.topwin, flags=gtk.DIALOG_MODAL|gtk.DIALOG_DESTROY_WITH_PARENT,
            buttons=(gtk.STOCK_OK, gtk.RESPONSE_OK, gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
        dialog.set_transient_for(self.topwin)
        table = gtk.Table(rows=2, columns=2)
        l_username = gtk.Label(vminstall.msg.username)
        table.attach(l_username,0,1,0,1)
        l_password = gtk.Label(vminstall.msg.password)
        table.attach(l_password,0,1,1,2)
        e_username = gtk.Entry()
        table.attach(e_username,1,2,0,1)
        e_password = gtk.Entry()
        e_password.set_visibility(False)
        table.attach(e_password,1,2,1,2)
        table.show_all()
        dialog.vbox.add(table)
        response = dialog.run()
        if response == gtk.RESPONSE_OK:
            smbuser = e_username.get_text()
            smbpasswd = e_password.get_text()
        else:
            smbuser = os.getenv('USER')
            smbpasswd = os.getenv('PASSWD')
        self.options.smboptions = "user=%s,passwd=%s" % (smbuser,smbpasswd)
        self.defaults.smboptions = self.options.smboptions
        dialog.destroy()
        log.debug("u:%s, p:%s", smbuser, smbpasswd)

def run(job, options, defaults):
    gtk.gdk.threads_init()
    assert(not options.background)
    window = VMCreate(job, options, defaults)
    window.topwin.connect("delete-event", gtk.main_quit)
    window.connect("vmmcreate-closing", gtk.main_quit)
    window.show()
    gtk.main()

