# Common code for all kvm guests
#
# Initially derived from virt-inst, Copyright 2005-2006 Red Hat, Inc.
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import os

from exceptions import *
from log import log
import paths
import util
import vmdisks
#import xen_hypervisor
import hypervisor
import keytable
import time
from guest import Guest

class KVMGuest(Guest):
#    def _get_name_xen(self):
#        return 'name="%s"\n' % self.options.name

    def _get_memory_xml(self):
        # XML wants KB
        return ['<memory>%d</memory>' % (self.options.max_memoryMB*1024,),
                '<currentMemory>%d</currentMemory>' % (self.options.memoryMB*1024,)]

    def _get_uuid_xml(self):
        return ['<uuid>%s</uuid>' % self.options.uuid]

    def _get_action(self, action):
        if self.isInstall:
            return 'destroy'
        else:
            if action == 'crash': return self.options.on_crash
            if action == 'poweroff': return self.options.on_poweroff
            if action == 'reboot': return self.options.on_reboot
    def _get_action_xml(self):
        return ['<on_poweroff>%s</on_poweroff>' % self._get_action('poweroff'),
                '<on_reboot>%s</on_reboot>' % self._get_action('reboot'),
                '<on_crash>%s</on_crash>' % self._get_action('crash')]

    def _get_vcpus_xml(self):
        return ['<vcpu>%d</vcpu>' % self.options.vcpus]

    def _get_localtime_xml(self):
        clock = []
        hpet = None
        if self.options.hpet is False:
            hpet = "<timer name='hpet' present='no'/>"
        if self.os().localtime():
            clock = ["<clock offset='localtime'>"]
        else:
            clock = ["<clock offset='utc'>"]
        if hpet:
            clock.append(hpet)
        clock.append("</clock>")
        return clock

    def getCommonXml(self):
        s = []
        s.extend(self._get_name_xml())
        s.extend(self._get_uuid_xml())
        s.extend(self._get_memory_xml())
        s.extend(self._get_vcpus_xml())
#BFR not needed?        s.extend(self._get_keymap_xml())
        return s

    def _get_disk_xml(self):
        """Get the disk config in the libvirt XML format"""
        available_cdroms = { 'hda': 0, 'hdb':0, 'hdc':0, 'hdd':0 }
        ret = []
        for disk in self.options.disks + self.tmpdisks:
            use_virtio = self.options.virtio and disk.virtio
            ret.extend(disk.get_xml_config(use_virtio, available_cdroms))
        return ret

    def _get_network_xml(self):
        """Get the network config in the libvirt XML format"""
        nics = []
        for n in self.options.nics:
            model = ""
            opts = []
            if n.bridge:
                nic = ["<interface type='bridge'>"]
                opts.append("<source bridge='%s'/>" % n.bridge)
            else:
                nic = ["<interface type='user'>"]
            if n.mac:
                opts.append("<mac address='%s'/>" % n.mac)
            if n.model:
                opts.append("<model type='%s'/>" % n.model)
#BFR            opts.append("<script path='/etc/xen/qemu-ifup'/>")
            nic.append(opts)
            nic.append('</interface>')
            nics.append(nic)
        return nics

    def _get_keymap_xml(self):
        ret = []
        keymap = keytable.get_keymap(self.options.keymap)
        if keymap != None:
            ret = ['<keymap>%s</keymap>' % keymap]
        return ret

    def _get_input_xml(self):
        """Get the input config in the libvirt XML format."""
        if self.options.graphics == 'none':
            return []
        if self.options.usb_mouse_abs:
            ret = "<input type='tablet' bus='usb'/>"
        else:
            ret = "<input type='mouse' bus='ps2'/>"
        return [ret]

    def _get_graphics_xml(self):
        """Get the graphics config in the libvirt XML format."""
        if self.options.graphics == 'none':
            return []
        ret = ''
        if self.options.graphics_viewer == 'vnc' or self.options.graphics_viewer == 'virt-viewer':
            ret = "<graphics type='vnc'"
            if self.options.vncport:
                ret += " port='%d'" %(self.options.vncport,)
            else:
                ret += " port='-1' autoport='yes'"
            if self.options.vnc_password:
                ret += " password='%s'" %(self.options.vnc_password,)
            if self.options.keymap:
                ret += " keymap='%s'" %(self.options.keymap,)
            ret += '/>'
        elif self.options.graphics_viewer == 'sdl':
            ret = "<graphics type='sdl'/>"
        return [ret]

    def GetConfigXml(self):
        raise RuntimeError('pure virtual')

    def PreInstall(self, progress=lambda f:None):
        if hypervisor.connection.domain_running(self.options.name):
            raise VmError(VmError.E_VM_EXISTS)
        if self.options.os_settings:
            self.os().actions().setupOsSettings(self, lambda f:progress(f*0.05))
        self.create_devices(lambda f:progress(0.05 + f*0.95))

    def Start(self):
        def indent(nodes, depth=0):
            def process(node):
                if type(node) is str:
                    return '  '*depth+node
                return indent(node, depth+1)
            return '\n'.join(map(process, nodes))
        if (hypervisor.connection.domain_running(self.options.name) or
            hypervisor.connection.domain_running(self.options.uuid)):
            raise VmError(VmError.E_VM_EXISTS)
        # If there is an existing configuration by this name it is not running,
        # but it may have a different (conflicting) UUID, so delete it.
        hypervisor.connection.delete(self.options.name)
        try:
            virDom = hypervisor.connection.getconn().defineXML(indent(self.GetConfigXml()))
            virDom.create()
        except RuntimeError, e:
            log.info("domain create failed")
            self.Stop()
            hypervisor.connection.delete(self.options.name)
            raise VmError(VmError.E_NO_START, details=e.args[0])
        if not hypervisor.connection.wait_domain(self.options.name):
            log.info("Domain failed to start")
            self.Stop()
            hypervisor.connection.delete(self.options.name)
            raise VmError(VmError.E_NO_START)
        self.uuid = self.options.uuid

    def Stop(self):
        # Slightly safer to stop by running-uuid, if it exists, rather than name.
        name = self.uuid
        if not name:
            name = self.options.name
        try:
            hypervisor.connection.destroy(name)
        except:
            pass
        self.uuid = None

    def PostInstallConf(self):
        """Perform any post-install cleanup and configuration."""
        self.isInstall = False
        self._cleanup_tmpdisks()
        self.os().actions().postInstall(self)

    def manage_disks(self):
        dms = vmdisks.diskmap.DiskMapStatic()
        for disk in self.options.disks:
            dms.manage_disk(disk)
        dms.flush()

    def unmanage_disks(self):
        dms = vmdisks.diskmap.DiskMapStatic()
        for disk in self.options.disks:
            try:
                dms.unmanage_disk(disk)
            except:
                # Possible that someone rm'd the disk list file...
                pass
        dms.flush()

    def start_from_disk(self):
        """Restart the guest from its disks."""
        def indent(nodes, depth=0):
            def process(node):
                if type(node) is str:
                    return '  '*depth+node
                return indent(node, depth+1)
            return '\n'.join(map(process, nodes))
        self.create_devices()
        # Ensure that the install settings are removed; overlaying settings
        # might break (specifically in the case of kernel vs. bootloader).
        # we need to delay long enough to ensure the networking scripts don't operate on the wrong interface
        # since the old and new tap network interfaces have the same names.
        time.sleep(2)
        hypervisor.connection.delete(self.options.name)
        try:
            virDom = hypervisor.connection.getconn().defineXML(indent(self.GetConfigXml()))
            virDom.create()
        except RuntimeError, e:
            log.info(" domain create failed")
            self.Stop()
            hypervisor.connection.delete(self.options.name)
            raise VmError(VmError.E_NO_START, details=e.args[0])
        if not hypervisor.connection.wait_domain(self.options.name):
            log.info("Domain failed to start")
            self.Stop()
            hypervisor.connection.delete(self.options.name)
            raise VmError(VmError.E_NO_START)
        self.uuid = self.options.uuid

    def create_devices(self, progress=lambda f:None):
        disks = self.options.disks + self.tmpdisks
        times = map(lambda d: d.setup_time(), disks)
        cum_times = [0.0]
        for i in range(len(times)):
            cum_times.append(cum_times[-1]+times[i])
        for i in range(len(disks)):
            disks[i].setup(lambda f: progress(cum_times[i]/cum_times[-1] + f*cum_times[i+1]/cum_times[-1]))

    def FailedInstall(self):
        log.info("Installation failed, so destroying VM")
        self.Stop()
        hypervisor.connection.delete(self.options.name)
        if self.options.preserve_on_error:
            log.info("Installation failed, not cleaning up disks because preserve_on_error is set")
        else:
            log.info("Installation failed, so cleaning up disks")
            self._cleanup_tmpdisks()
            self.unmanage_disks()
            for disk in self.options.disks:
                disk.reset()

    def _cleanup_tmpdisks(self):
        for disk in self.tmpdisks:
            try:
                os.unlink(disk.get_path())
            except:
                pass
        self.tmpdisks = []

    def __del__(self):
        log.debug("KVMGuest.__del__")
        self._cleanup_tmpdisks()

