# Script to set up a Xen guest and kick off an install
#
# Initially derived from virt-manager, Copyright 2005-2006 Red Hat, Inc.
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import os
import sys

from vminstall.VMDefaults import VMDefaults as VMDefaults
from vminstall.nics import XenNic
import vmdisks
import vminstall
import vminstall.console
import vminstall.options
import vminstall.hypervisor

options = None
defaults = None

def output(s):
    if not isinstance(s, str):
        s = str(s)
    sys.stdout.write(s)

def output_wrap(s):
    output(vminstall.util.wrap(s))

def error(s):
    if not isinstance(s, str):
        s = str(s)
    vminstall.log.error(s)
    output_wrap('%s: %s\n' % (vminstall.msg.error, s))

def bad_noninter_default(default):
    raise vminstall.ParamError(err=vminstall.ParamError.E_INVALID, details=default)

def yes_or_no(s):
    if isinstance(s, bool):
        return s
    if s.lower() in ('y', 'yes'):
        return True
    elif s.lower() in ('n', 'no'):
        return False
    raise RuntimeError

def prompt_for_input(prompt=''):
    if options.background:
        raise vminstall.ParamError(err=vminstall.ParamError.E_INVALID)
    if prompt:
        prompt += ' '
    output(prompt + '> ')
    value = sys.stdin.readline().strip()
    if value == 'q' or len(value) == 1 and ord(value) == 27:
        sys.exit(0)
    return value

def prompt_string(prompt, default=None):
    if options.background:
        return default
    if default is not None:
        prompt += '[%s]' % default
    val = prompt_for_input(prompt)
    if val == '' and default is not None:
        val = default
    return val

def prompt_numeric(prompt, minimum=None, maximum=None, default=None, float_ok=False):
    if default is not None:
        if options.background:
            return default
        if prompt:
            prompt += ' '
        prompt += '[%s]' % str(default)
    while True:
        val = prompt_for_input(prompt)
        if val == '' and default is not None:
            val = default
            default = None
        else:
            try:
                val = int(val)
            except ValueError:
                if float_ok:
                    try:
                        val = float(val)
                    except ValueError:
                        error(vminstall.msg.prompt_number)
                        continue
                else:
                    error(vminstall.msg.prompt_number)
                    continue
            except:
                error(vminstall.msg.prompt_number)
                continue
        if minimum is not None and val < minimum:
            error(vminstall.msg.min_val % {'val':str(minimum)})
        elif maximum is not None and val > maximum:
            error(vminstall.msg.max_val % {'val':str(maximum)})
        else:
            return val

def prompt_yes_no(prompt, default=None):
    if default is None:
        default_str = ''
    else:
        assert(isinstance(default, bool))
        if options.background:
            return default
        default_str = '[%s]'
        if default:
            default_str %= 'Y'
        else:
            default_str %= 'N'
    if prompt:
        prompt += ' '
    while True:
        val = prompt_for_input('%s(Y / N) %s' % (prompt, default_str))
        if val == '' and default is not None:
            return default
        try:
            return yes_or_no(val)
        except RuntimeError:
            continue

def prompt_selection_by_key(title, prompt, selections, default=None):
    """Prompts the user to make a selection from a list by typing a key.
    'selections' is a dictionary mapping from key to displayable value.
    The typed key is returned."""
    keys = vminstall.util.sort_dict_keys_by_value(selections)
    if options.background:
        if not default in keys:
            bad_noninter_default(default)
        return default
    def longest(a, b):
        if len(a) > len(b):
            return a
        else:
            return b
    maxlen = len(reduce(longest, keys))
    while True:
        if title:
            output(vminstall.util.wrap(title) + '\n')
        for key in keys:
            output('  %s%s: %s\n' % (' ' * (maxlen - len(key)), key, selections[key]))
        val = prompt_string(prompt, default)
        if not val in keys:
            output('\n')
        else:
            return val

def prompt_selection_by_number(title, selections, default=None):
    """Prompts the user to make a selection from a list by typing a number.
    'selections' is a dictionary mapping from key to displayable value.
    The key cooresponding to the user's selection is returned."""
    keys = vminstall.util.sort_dict_keys_by_value(selections)
    if options.background:
        if not default in keys:
            bad_noninter_default(default)
        return default
    while True:
        indicies = {}
        i = 1
        default_i = None
        if title:
            output(vminstall.util.wrap(title) + '\n')
        for key in keys:
            output(' %2d: %s\n' % (i, selections[key]))
            indicies[i] = key
            if default == key:
                default_i = i
            i += 1
        val = prompt_numeric('', 1, len(indicies), default_i)
        return indicies[val]

# General input gathering functions
def get_os_type():
    def getPrettyName(name):
        return (name, VMDefaults[name].name())
    osnames = dict(map(getPrettyName, VMDefaults))
    if options.os_type is not None and osnames.has_key(options.os_type):
        return
    if options.background and osnames.has_key(defaults.os_type):
        options.os_type = defaults.os_type
        return
    help_text = vminstall.msg.os_type_help + '\n' + vminstall.msg.cli_exit_help
    options.os_type = prompt_selection_by_number(help_text, osnames, defaults.os_type)
    output('\n')

def get_virt_method():
    selections = {}
    if options.para_possible():
        selections['p'] = vminstall.msg.paravirtualization
    if options.full_possible():
        selections['f'] = vminstall.msg.full_virtualization
    if len(selections) > 1:
        default_method = 'p'
        if defaults.full_virt:
            default_method = 'f'
        full_virt = (prompt_selection_by_number(vminstall.msg.method_help,
                                               selections, default_method) == 'f')
        output('\n')
    else:
        full_virt = defaults.full_virt
    options.full_virt = full_virt
    if options.full_virt:
        output(vminstall.msg.PXE_boot + '\n')
        if prompt_yes_no('', False):
            options.pxe_boot = True
            output('\n')

def get_name():
    if options.name:
        return
    defaultName = defaults.name
    output_wrap(vminstall.msg.name_help + '\n')
    while True:
        name = prompt_string('', defaultName)
        try:
            options.name = name
            break
        except vminstall.Error, e:
            error(e)
            if name == defaultName:
                defaultName = None
    output('\n')

def get_description():
    if options.description:
        return
    defaultDescription = defaults.description
    while True:
        desc = prompt_string(vminstall.msg.description, defaultDescription)
        try:
            options.description = desc
            break
        except vminstall.Error, e:
            error(e)
            if desc == defaultDescription:
                defaultDescription = None
    output('\n')

def get_hardware():
    def get_num(prompt, minimum, maximum, default, name):
        while True:
            value = prompt_numeric(prompt, minimum, maximum, default)
            try:
                options.__dict__[name] = value
                break
            except vminstall.Error, e:
                error(e)
                if value == default:
                    default = None

    if (options.memoryMB and
        options.max_memoryMB and
        options.vcpus):
        return
    output_wrap(vminstall.msg.hardware_help + '\n')
    vmdefaults = VMDefaults[options.os_type]
    # Calculate initial and max memory based on minimum initial 
    # for this type of guest
    init_memory, max_memory = vminstall.options.calculate_max_memory(options, vmdefaults.min_memory(), vmdefaults.max_memory())
    if not options.memoryMB:
        get_num(vminstall.msg.memory_prompt, vmdefaults.min_memory(),
                init_memory, defaults.memoryMB, 'memoryMB')
    # Re-calculate max memory based on new initial memory
    init_memory, max_memory = vminstall.options.calculate_max_memory(options, options.memoryMB, vmdefaults.max_memory())
    if not options.max_memoryMB:
        get_num(vminstall.msg.max_memory_prompt, options.memoryMB,
                max_memory, options.memoryMB, 'max_memoryMB')
    if options.max_memoryMB > options.memoryMB:
        output_wrap(vminstall.msg.warning + ': ' + vminstall.msg.maximum_memory_warning + '\n')
    if not options.vcpus:
        get_num(vminstall.msg.vcpus_prompt, 1,
                vmdefaults.max_cpus(), defaults.vcpus, 'vcpus')
    output('\n')

def get_disks():
    if options.disks:
        default_disks = []
    else:
        # options.disks starts out as None to indicate that no choice has
        # been made.  For ease of coding, now change to empty list.
        options.disks = []
        default_disks = defaults.disks

    virtual_disks = vmdisks.get_possible_vdevs(options.full_virt, options.virtio)

    mask_default = False
    while len(options.disks) < len(virtual_disks):
        output(vminstall.msg.virtual_disks + ':\n')
        if len(options.disks) == 0:
            output('  (%s)\n' % vminstall.msg.none)
        else:
            for disk in options.disks:
                output('  %s\n' % str(disk))
        output(vminstall.msg.another_disk + '\n')
        if not prompt_yes_no('', (len(options.disks) == 0)):
            output('\n')
            break

        output_wrap('\n' + vminstall.msg.disk_help + '\n')

        default_disk = None
        if (not options.background or not mask_default) and \
           len(options.disks) < len(default_disks):
            default_disk = default_disks[len(options.disks)]
        mask_default = False

        # Virtual type (disk, cdrom, ...)
        if default_disk:
            default_vtype = default_disk._vdevType[0]
        else:
            default_vtype = vmdisks.Disk.DEVICE_DISK[0]
        vtype_selections = {
            'd': vminstall.msg.type_hard_disk,
            'c': vminstall.msg.type_cdrom,
        }
        if options.full_virt:
            vtype_selections['f'] = "Floppy"
        vtype = prompt_selection_by_number(
            vminstall.msg.add_disk_type, vtype_selections, default_vtype)
        if vtype == 'd':
            vtype = vmdisks.Disk.DEVICE_DISK
        elif vtype == 'c':
            vtype = vmdisks.Disk.DEVICE_CDROM
        else:
            vtype = vmdisks.Disk.DEVICE_FLOPPY

        output(vminstall.msg.add_disk_source + '\n')
        default = '%s%s/%s' % (vminstall.paths.imagedir, options.name,
                               virtual_disks[len(options.disks)])
        path = prompt_string('', default)
        sizeMB = 0
        useSparse = True
        if not os.path.exists(os.path.abspath(path)):
            default_sizeGB = None
            if default_disk:
                default_sizeGB = default_disk.get_virtual_size() / (2.0*1024*1024)
            sizeGB = prompt_numeric(vminstall.msg.size_gb,
                1, 1024, default_sizeGB, float_ok=True)
            sizeMB = int(sizeGB * 1024)
            output(vminstall.msg.create_sparse_file + '\n')
            if not prompt_yes_no('', True):
                useSparse = False
                output('\n')

        try:
            disk = vmdisks.Disk(path, virtual_disks[len(options.disks)],
                                blocks=sizeMB*1024*2, vdevType=vtype)
            disk.args['sparse'] = useSparse
        except vminstall.Error, e:
            mask_default = True
            error(e)
            output('\n')
            continue
        options.disks.append(disk)
        output('\n')

def get_network():
    if options.nics:
        default_nics = []
    else:
        # options.nics starts out as None to indicate that no choice has
        # been made.  For ease of coding, now change to empty list.
        options.nics = []
        default_nics = VMDefaults[options.os_type].nics()
        for nic in default_nics:
            nic.update(options.full_virt)

    while len(options.nics) < vminstall.hypervisor.max_nics:
        output(vminstall.msg.title_nics + '\n')
        if len(options.nics) == 0:
            output('  (%s)\n' % vminstall.msg.none)
        else:
            for nic in options.nics:
                output('  %s\n' % str(nic))
        output(vminstall.msg.another_nic + '\n')
        if not prompt_yes_no('', (len(options.nics) == 0)):
            output('\n')
            break

        default_nic = { 'mac': None, 'model': 'para' }
        if len(options.nics) < len(default_nics):
            default_nic = default_nics[len(options.nics)]

        if options.full_virt:
            model = prompt_selection_by_number(vminstall.msg.add_nic_type,
                                               vminstall.nics.XenNic.models, 'rtl8139')
            nic = vminstall.nics.XenNic(full_virt=options.full_virt, model=model)
        else:
            nic = vminstall.nics.XenNic(full_virt=options.full_virt)
        options.nics.append(nic)
        output('\n')

def get_graphics():
    if options.graphics:
        return

    output_wrap(vminstall.msg.graphics_hw_help + '\n')
    grhw = {'none': vminstall.msg.graphics_hw_none}
    if options.full_virt:
        grhw['cirrus'] = vminstall.msg.graphics_hw_cirrus
        grhw['vesa'] = vminstall.msg.graphics_hw_vesa
        default = 'cirrus'
    else:
        grhw['para'] = vminstall.msg.graphics_hw_para
        default = 'para'
    options.graphics = prompt_selection_by_number('', grhw, default)
    output('\n')

def _get_pv_install():
    if options.source:
        return

    vmdefaults = VMDefaults[options.os_type]
    install_selections = {}
    default = None
    protocols = vmdefaults.install_sources()[1]
    network_protocols = []
    for p in protocols:
        if p == 'dev:/':
            for disk in options.disks:
                if disk.exists():
                    url = 'dev:/%s' % (disk.vdev,)
                    install_selections[url] = '%s (%s)' % (vminstall.msg.virtual_disk , str(disk),)
                    if default is None:
                        default = url
        else:
            network_protocols.append(p)
    if network_protocols:
        default = 'network'
        install_selections[default] = '%s (%s)' % (vminstall.msg.network_url, ', '.join(network_protocols),)
    if default is None:
        # Validation step will catch this error, and allow user to retry.
        return

    while True:
        url = prompt_selection_by_number(vminstall.msg.install_media_help, install_selections, default)

        try:
            if not url.startswith('dev:/'):
                url = prompt_string("Type the install URL:")
            options.source = url
            break
        except vminstall.InstSrcError, e:
            error(e)
    output('\n')

def _get_pv_os_settings():
    if options.os_settings:
        return
    vmdefaults = VMDefaults[options.os_type]
    auto_install_type = vmdefaults.auto_install()
    if not auto_install_type:
        return
    output_wrap(vminstall.msg.os_settings_help + '\n')
    while True:
        os_settings = prompt_string(auto_install_type)
        try:
            options.os_settings = os_settings
            break
        except vminstall.Error, e:
            error(e)
    output('\n')


def _get_pv_extra_args():
    if options.extra_args:
        return
    if not options.extra_args:
        output_wrap(vminstall.msg.extra_args_help + '\n')
        vmdefaults = VMDefaults[options.os_type]
        options.extra_args = prompt_string(vminstall.msg.extra_args_prompt)
        output('\n')

def _get_fv_install():
    # Anything to do here?  Maybe as "boot c" vs "boot d"?
    pass

def get_install_source():
    if options.full_virt:
        _get_fv_install()
    else:
        _get_pv_install()
        _get_pv_os_settings()
        _get_pv_extra_args()

def _getSettingsCB(job):
    while True:
        if job.upgrade:
            if defaults.upgrade_managed_vm:
                vm_name =  defaults.name
            else:
                vm_name = defaults.upgrade_config
            vminstall.options.initialize_vm_from_config_file(defaults, vm_name)
        else:
            for fn in (#get_install_or_create,
                       get_os_type,
                       get_virt_method,
                       get_name,
                       get_description,
                       get_hardware,
                       get_graphics,
                       get_disks,
                       get_network,
                       get_install_source):
                # Get the user's input...
                fn()
                # ...and recalculate all remaining defaults
                # FIXME: This can throw.  But how to back up and undo?
                #        --> similar to gui's save_page
                vminstall.options.calculate(options, defaults)

        try:
            vminstall.options.finalize(defaults)
            break
        except vminstall.Error, e:
            if options.background:
                raise
            error(e)
            output('%s\n%s\n\n' % (vminstall.msg.correct_retry, '=' * 40))

    return defaults

def _stateChangeCB(job, state):
    if state[0] == job.POST_INSTALL:
        vminstall.console.reset()
    output('%s\n\n' % state[2])

def _errorCB(job, text):
    output('%s\n' % text)

def run(job, _options, _defaults):
    global options
    options = _options
    global defaults
    defaults = _defaults

    job.run(GetSettingsCB=_getSettingsCB,
            StateChangeCB=_stateChangeCB,
            ErrorCB=_errorCB)
